/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "CMDLineRegistry.h"
#include "CMDLineHelpProvider.h"

#include <cstdio>

namespace GB2 {

static bool isDoubleDashParam(const QString& val) {
    return val.startsWith("--") && val.length() >= 3 && val.at(2).isLetter();;
}

static bool isSingleDashParam(const QString& val) {
    return val.startsWith("-") && val.length() >= 2 && val.at(1).isLetter();
}


static bool tryParseDoubleDashParameter(const QString& argument, QString& paramName, QString &paramValue) {
    if (!isDoubleDashParam(argument)) {
        return false;
    }
    int nameEndIdx = argument.indexOf("=");
    if (nameEndIdx == -1) {
        paramName = argument.mid(2);
    } else {
        paramName = argument.mid(2, nameEndIdx - 2);
        paramValue = argument.mid(nameEndIdx+1);
    }
    return true;
}

static bool tryParseSingleDashParameter(const QString& argument, const QString& nextArgument, QString& paramName, QString &paramValue) {
    if (!isSingleDashParam(argument)) {
        return false;
    }
    paramName = argument.mid(1);
    if (!isDoubleDashParam(nextArgument) && !isSingleDashParam(nextArgument)) {
        paramValue = nextArgument;    
    }    
    return true;
}


CMDLineRegistry::CMDLineRegistry(const QStringList& arguments) {
    for(int i=0; i<arguments.size(); i++) {
        const QString& argument = arguments.at(i);
        QString paramName, paramValue;
        if (!tryParseDoubleDashParameter(argument, paramName, paramValue)) {
            QString nextArgument;
            if (i < arguments.size() - 1) {
                nextArgument = arguments.at(i + 1);
            }
            if (tryParseSingleDashParameter(argument, nextArgument, paramName, paramValue)) {
                if (!paramValue.isEmpty()) {
                    i++;
                }
            } else {
                paramName = argument;
            }
        }
        orderedParamNames.append(paramName);
        params[paramName] = paramValue;
    }
}

static bool providerNameComparator(const CMDLineHelpProvider* p1, const CMDLineHelpProvider* p2) {
    return p1->getHelpSectionName().compare(p2->getHelpSectionName()) > 0;
}

void CMDLineRegistry::registerCMDLineHelpProvider(CMDLineHelpProvider* provider) {
    helpProviders.append(provider);
    qStableSort(helpProviders.begin(), helpProviders.end(), providerNameComparator); 
}

void CMDLineRegistry::unregisterCMDLineHelpProvider(CMDLineHelpProvider* provider) {
    helpProviders.removeOne(provider);
}

void CMDLineRegistry::dumpHelp() const {
    QString prevSectionName;
    foreach (CMDLineHelpProvider* hProvider, helpProviders) {
        const QString& sectionName = hProvider->getHelpSectionName();
        if ( sectionName != prevSectionName) {
            fprintf(stdout, QString("Section: " + sectionName + "\n").toLocal8Bit().constData());
            prevSectionName = sectionName;
        }
        fprintf(stdout, hProvider->getHelpSectionContent().toLocal8Bit().constData());
        fprintf(stdout, "\n");
    }
}

} //namespace
