/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "MAlignment.h"
#include <datatype/DNASequence.h>

#include <limits.h>

namespace GB2 {

MAlignment::MAlignment(const QString& _name, DNAAlphabet* al, const QList<MAlignmentItem>& aitems) 
: alphabet(al), alignedSeqs(aitems)
{
    assert(al==NULL || !_name.isEmpty()); //if not default constructor -> name must be provided
    setName( _name );
    normalizeModel();
}

void MAlignment::normalizeModel() {
    //check that all sequences are of equal size
    int maxSize = 0;
    for (int i=0, n = alignedSeqs.size();  i < n; i++) {
        const MAlignmentItem& item = alignedSeqs[i];
        maxSize = qMax(item.sequence.size(), maxSize);
    }

    for (int i=0, n = alignedSeqs.size();  i < n; i++) {
        MAlignmentItem& item = alignedSeqs[i];
        int dSize = maxSize - item.sequence.size();
        if (dSize > 0) {
            item.sequence.append(QByteArray(dSize, MAlignment_GapChar));
        }
    }
}


bool MAlignment::isNormalized() const {
    if (alphabet == NULL) {
        return false;
    }
    //check that all sequences are of equal size
    int size = 0;
    for (int i=0, n = alignedSeqs.size();  i < n; i++) {
        const MAlignmentItem& item = alignedSeqs[i];
        if (i == 0) {
            size = item.sequence.size();
        } else {
            int itemSize = item.sequence.size();
            if (size!=itemSize) {
                return false;
            }
        }
    }

    //todo: check alphabet too

    return true;
}

bool MAlignment::hasGaps() const {
    for (int i=0, n = alignedSeqs.size();  i < n; i++) {
        const MAlignmentItem& item = alignedSeqs[i];
        if (item.sequence.indexOf(MAlignment_GapChar)!=-1) {
            return true;
        }
    }
    return false;
}

char MAlignment::getBase(int seqNum, int pos) const {
    assert(seqNum >= 0 && seqNum < getNumSequences());
    assert(pos >=0 && pos < getLength());
    const MAlignmentItem& item = alignedSeqs[seqNum];
    return item.sequence[pos];
}


void MAlignment::clear() {
    alphabet = NULL;
    alignedSeqs.clear();
}


MAlignment MAlignment::subAlignment(int start, int len) const {
    assert(start >=0 && start + len <= getLength());
    assert(isNormalized());

    MAlignment res;
    res.alphabet = alphabet;
    foreach(const MAlignmentItem& item, alignedSeqs) {
        res.alignedSeqs.append(MAlignmentItem(item.name, item.sequence.mid(start, len)));
    }
    return res;
}


MAlignment& MAlignment::operator+=(const MAlignment& ma) {
    assert(ma.alphabet == alphabet);
    assert(isNormalized());
    assert(ma.isNormalized());

    int nSeq = getNumSequences();
    assert(ma.getNumSequences() == nSeq);
    
    for (int i=0; i<nSeq; i++) {
        MAlignmentItem& item = alignedSeqs[i];
        item.sequence.append(ma.alignedSeqs[i].sequence);
    }
    return *this;
}

int MAlignment::estimateMemorySize() const {
    int len = getLength();
    int nSeq = getNumSequences();
    int res = len * nSeq + 100 * nSeq;
    return res;
}

bool MAlignment::makeSubAlignment( LRegion window, QStringList seqNames ){
    QList<MAlignmentItem> newList, oldList = alignedSeqs;
    foreach(MAlignmentItem item, alignedSeqs){
        if(seqNames.contains(item.name)){
            MAlignmentItem newItem(item.name);
            newItem.sequence = item.sequence.mid(window.startPos, window.len);
            newList.append(newItem);
        }
    }
	alignedSeqs = newList;
	if(!removeGaps()){
		alignedSeqs = oldList;
		return false;
	}    
	return true;
}

bool MAlignmentItem::operator!=( MAlignmentItem &ob ){
	return (this->name != ob.name) || (this->sequence != ob.sequence);
}

bool MAlignment::removeGaps(){
	int commonGapsLenFromStart = INT_MAX, commonGapsLenFromEnd = INT_MAX;
	for(QList<MAlignmentItem>::iterator it = alignedSeqs.begin(); it != alignedSeqs.end(); it++){
		int currentGapsFromStart = 0, currentGapsFromEnd = 0;
		detectGapsLen(*it, currentGapsFromStart, currentGapsFromEnd);
		commonGapsLenFromStart = (commonGapsLenFromStart > currentGapsFromStart) ? currentGapsFromStart : commonGapsLenFromStart;
		commonGapsLenFromEnd = (commonGapsLenFromEnd > currentGapsFromEnd) ? currentGapsFromEnd : commonGapsLenFromEnd;
	}
	if(commonGapsLenFromEnd == alignedSeqs.first().sequence.length()){
		return false;
	}
	for(QList<MAlignmentItem>::iterator it = alignedSeqs.begin(); it != alignedSeqs.end(); it++){
		int currentGapsFromStart = 0, currentGapsFromEnd = 0;
		detectGapsLen(*it, currentGapsFromStart, currentGapsFromEnd);
		commonGapsLenFromStart = (commonGapsLenFromStart > currentGapsFromStart) ? currentGapsFromStart : commonGapsLenFromStart;
		commonGapsLenFromEnd = (commonGapsLenFromEnd > currentGapsFromEnd) ? currentGapsFromEnd : commonGapsLenFromEnd;
	}

	for(QList<MAlignmentItem>::iterator it = alignedSeqs.begin(); it != alignedSeqs.end(); it++){
		MAlignmentItem *itm = &(*it);
		itm->sequence = itm->sequence.left(itm->sequence.length() - commonGapsLenFromEnd);
		itm->sequence = itm->sequence.right(itm->sequence.length() - commonGapsLenFromStart);
	}

	return true;
}

void MAlignment::detectGapsLen(MAlignmentItem &item, int &fromStart, int &fromEnd){
	QByteArray arr = item.sequence;
	for(int i = 0; i != arr.length(); i++){
		if(arr.at(i) == MAlignment_GapChar){
			fromStart++;
		}else break;
	}

	for(int i = arr.length()-1; i != -1; i--){
		if(arr.at(i) == MAlignment_GapChar){
			fromEnd++;
		}else break;
	}
}

bool MAlignment::operator!=( const MAlignment& other ) const
{
    if(alignedSeqs.size() != other.alignedSeqs.size()){
        return true;
    }
    for(int i = 0; i < alignedSeqs.size(); i++){
        MAlignmentItem actItem = alignedSeqs.at(i), expItem = other.alignedSeqs.at(i);
        //printf("\n%s\n", actItem.sequence.constData());
        //printf("\n%s\n", expItem.sequence.constData());

        if (actItem != expItem){
            return true;
        }
    }

    return false;

}

void MAlignment::addSequence( const DNASequence& seq, int seqIdx, bool normailize )
{
    assert(isNormalized());
    
    MAlignmentItem mai(seq.getName(), seq.seq);
    if (seqIdx < 0 || seqIdx >= getLength()) {
        alignedSeqs.append(mai);
    } else {
        alignedSeqs.insert(seqIdx, mai);
    }

    if (normailize) {
        normalizeModel();
    }
        
}

}//namespace
