/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/


#ifndef _GB2_SERIALIZE_UTILS_H_
#define _GB2_SERIALIZE_UTILS_H_

#include <core_api/core_api.h>
#include <core_api/AppContext.h>
#include <core_api/DNAAlphabet.h>
#include <datatype/DNASequence.h>
#include <datatype/MAlignment.h>
#include <distributed_computing/RemoteMachine.h>
#include <core_api/VirtualFileSystem.h>

namespace GB2 {

class GB2_COREAPI_EXPORT SerializeUtils {
private:
    static const int SERIALIZED_DNASEQUENCE_LIST_SZ = 3;
    static const int SERIALIZED_MALIGNMENT_LIST_SZ = 3;
    
    static const QString REMOTE_MACHINE_VERSION_1_0;
    
public:
    template<typename T> static QVariant serializeValue( const T & t ) {
        return t;
    }
    
    template<typename T> static bool deserializeValue( const QVariant & data, T * val ) {
        if( NULL == val ) {
            return false;
        }
        if( !data.canConvert<T>() ) {
            return false;
        }
        *val = data.value<T>();
        return true;
    }
    
    template<typename T> static QVariant serializeArray( T * arr, int sz ) {
        if( NULL == arr ) {
            return 0;
        }
        assert( 0 <= sz );
        QVariantList res;
        for( int i = 0; i < sz; ++i ) {
            res << serializeValue<T>( arr[i] );
        }
        return res;
    }
    
    /* the arr array must be allocated by caller */
    template<typename T> static bool deserializeArray( const QVariant & data, T * arr, int sz ) {
        if( data.canConvert( QVariant::Int ) && 0 == data.toInt() ) {
            return true;
        }
        if( 0 >= sz ) {
            return false;
        }
        if( !data.canConvert( QVariant::List ) ) {
            return false;
        }
        QVariantList args = data.toList();
        if( args.size() != sz ) {
            return false;
        }
        
        for( int i = 0; i < sz; ++i ) {
            if( !deserializeValue<T>( args[i], arr + i ) ) {
                return false;
            }
        }
        return true;
    }
    
    /* serialization of RemoteMachineSettings: human-readable QString instead of binary QVariant 
       3 deserialization functions need for tests */
    static QString serializeRemoteMachineSettings( RemoteMachineSettings * machine );
    static bool deserializeRemoteMachineSettings( const QString & data, RemoteMachineSettings ** settings, QString * protoId = NULL );
    static bool deserializeRemoteMachineSettings( const QString & data, RemoteMachine ** machine );
    static bool deserializeRemoteMachineSettingsFromFile( const QString & machinePath, RemoteMachine ** machine );
    
}; // SerializeUtils

// serializeValue/deserializeValue template realizations: TODO: is it good to make them inline?
template<>
inline QVariant SerializeUtils::serializeValue<char*>( char * const& str ) {
    QVariantList res;
    if( NULL == str ) {
        res << serializeValue<int>( -1 );
        res << QVariant();
    } else {
        int len = strlen( str );
        res << serializeValue<int>( len );
        res << serializeArray<char>( str, len );
    }
    return res;
}

template<>
inline bool SerializeUtils::deserializeValue<char*>( const QVariant & data, char ** str ) {
    if( NULL == str ) {
        return false;
    }
    *str = NULL;

    if( !data.canConvert( QVariant::List ) ) {
        return false;
    }
    QVariantList args = data.toList();
    if( 2 != args.size() ) {
        return false;
    }

    int sz = 0;
    if( !deserializeValue<int>( args[0], &sz ) ) { return false; }
    if( -1 == sz ) {
        return true;
    }
    assert( 0 <= sz );
    *str = (char*)calloc( sz + 1, sizeof( char ) );
    assert( NULL != str );
    if( !deserializeArray<char>( args[1], *str, sz ) ) { return false; }
    (*str)[sz] = '\0';
    return true;
}

template<>
inline QVariant SerializeUtils::serializeValue<DNAAlphabet*>( DNAAlphabet * const& al ) {
    QVariantList res;
    if( NULL != al ) {
        res << serializeValue<bool>( true );
        res << serializeValue<QString>( al->getId() );
    } else {
        res << serializeValue<bool>( false );
        res << serializeValue<QString>( "" );
    }
    return res;
}

template<>
inline bool SerializeUtils::deserializeValue<DNAAlphabet*>( const QVariant & data, DNAAlphabet ** al ) {
    if( NULL == al ) {
        return false;
    }
    *al = NULL;
    if( !data.canConvert( QVariant::List ) ) {
        return false;
    }

    QVariantList args = data.toList();
    if( 2 != args.size() ) {
        return false;
    }

    bool notNullAl = false;
    QString alId;
    if( !deserializeValue<bool>( args[0], &notNullAl ) ) { return false; }
    if( !deserializeValue<QString>( args[1], &alId ) ) { return false; }
    if( notNullAl ) {
        DNAAlphabetRegistry * alReg = AppContext::getDNAAlphabetRegistry();
        assert( NULL != alReg );
        *al = alReg->findById( alId );
    }

    return true;
}

template<>
inline QVariant SerializeUtils::serializeValue<DNASequence>( const DNASequence & sequence ) {
    assert( !sequence.isNull() && NULL != sequence.alphabet );
    QVariantList res;
    res << serializeValue<QVariantMap>( sequence.info );
    res << serializeValue<QByteArray>( sequence.seq );
    res << serializeValue<DNAAlphabet*>( sequence.alphabet );
    return res;
}

template<>
inline bool SerializeUtils::deserializeValue<DNASequence>( const QVariant & data, DNASequence * seq ) {
    if( NULL == seq ) {
        return false;
    }
    if( !data.canConvert( QVariant::List ) ) {
        return false;
    }
    QVariantList args = data.toList();
    if( SERIALIZED_DNASEQUENCE_LIST_SZ != args.size() ) {
        return false;
    }

    if( !deserializeValue<QVariantMap>( args[0], &seq->info ) ) { return false; }
    if( !deserializeValue<QByteArray>( args[1], &seq->seq ) ) { return false; }
    if( !deserializeValue<DNAAlphabet*>( args[2], &seq->alphabet ) ) { return false; }
    assert( NULL != seq->alphabet );

    return true;
}

template<>
inline QVariant SerializeUtils::serializeValue<MAlignmentItem>( const MAlignmentItem & it ) {
    QVariantList res;
    res << serializeValue<QString>( it.name );
    res << serializeValue<QByteArray>( it.sequence );
    return res;
}

template<>
inline QVariant SerializeUtils::serializeValue< QList<MAlignmentItem> >( const QList< MAlignmentItem > & items ) {
    QVariantList res;
    foreach( const MAlignmentItem & item, items ) {
        res << serializeValue<MAlignmentItem>( item );
    }
    return res;
}

template<>
inline bool SerializeUtils::deserializeValue<MAlignmentItem>( const QVariant & data, MAlignmentItem * item ) {
    if( !data.canConvert( QVariant::List ) ) {
        return false;
    }
    QVariantList args = data.toList();
    if( 2 != args.size() ) {
        return false;
    }

    if( !deserializeValue<QString>( args[0], &item->name ) ) { return false; }
    if( !deserializeValue<QByteArray>( args[1], &item->sequence ) ) { return false; }
    return true;
}

template<>
inline bool SerializeUtils::deserializeValue< QList<MAlignmentItem> >( const QVariant & data, QList< MAlignmentItem > * itemList ) {
    if( !data.canConvert( QVariant::List ) ) {
        return false;
    }

    QVariantList args = data.toList();
    foreach( const QVariant & arg, args ) {
        MAlignmentItem item;
        if( !deserializeValue<MAlignmentItem>( arg, &item ) ) { return false; }
        itemList->append( item );
    }
    return true;
}

template<>
inline QVariant SerializeUtils::serializeValue<MAlignment>( const MAlignment & ma ) {
    QVariantList res;
    res << serializeValue<DNAAlphabet*>( ma.alphabet );
    res << serializeValue<QVariantMap>( ma.info );
    res << serializeValue< QList<MAlignmentItem> >( ma.alignedSeqs );
    return res;
}

template<>
inline bool SerializeUtils::deserializeValue<MAlignment>( const QVariant & data, MAlignment * ma ) {
    if( NULL == ma ) {
        return false;
    }
    if( !data.canConvert( QVariant::List ) ) {
        return false;
    }
    QVariantList args = data.toList();
    if( SERIALIZED_MALIGNMENT_LIST_SZ != args.size() ) {
        return false;
    }
    
    if( !deserializeValue<DNAAlphabet*>( args[0], &ma->alphabet ) ) { return false; }
    if( !deserializeValue<QVariantMap>( args[1], &ma->info ) ) { return false; }
    if( !deserializeValue< QList<MAlignmentItem> >( args[2], &ma->alignedSeqs ) ) { return false; }
    return true;
}

template<>
inline QVariant SerializeUtils::serializeValue<VirtualFileSystem>( const VirtualFileSystem & vfs ) {
    QVariantList res;
    
    res << serializeValue( vfs.getId() );
    
    QVariantMap resMap;
    QStringList filenames = vfs.getAllFilenames();
    foreach( const QString & filename, filenames ) {
        QByteArray file = vfs.getFileByName( filename );
        resMap.insert( filename, serializeValue( file ) );
    }
    res << serializeValue( resMap );
    
    return res;
}

template<>
inline bool SerializeUtils::deserializeValue<VirtualFileSystem>( const QVariant & data, VirtualFileSystem * val ) {
    if( NULL == val ) {
        return false;
    }
    if( !data.canConvert( QVariant::List ) ) {
        return false;
    }
    QVariantList args = data.toList();
    if( 2 != args.size() ) {
        return false;
    }
    
    QString id;
    if( !deserializeValue( args[0], &id ) ) { return false; }
    val->setId( id );
    
    QVariantMap filesMap;
    if( !deserializeValue( args[1], &filesMap ) ) { return false; }
    foreach( const QString & filename, filesMap.keys() ) {
        QByteArray fileData;
        if( !deserializeValue( filesMap[filename], &fileData ) ) {
            val->removeAllFiles();
            return false;
        }
        val->createFile( filename, fileData );
    }
    
    return true;
}

} // GB2

#endif // _GB2_SERIALIZE_UTILS_H_
