/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "GFFFormat.h"
#include "DocumentFormatUtils.h"

#include <core_api/IOAdapter.h>
#include <core_api/Task.h>
#include <util_gui/GUIUtils.h>
#include <gobjects/AnnotationTableObject.h>
#include <util_text/TextUtils.h>
#include <core_api/L10n.h>
#include <util_algorithm/GAutoDeleteList.h>


namespace GB2{

#define READ_BUFF_SIZE 4096

GFFFormat::GFFFormat(QObject* p):DocumentFormat(p, DocumentFormatFlag(0), QStringList("gff")){
    formatName = tr("GFF");
    supportedObjectTypes+=GObjectTypes::ANNOTATION_TABLE;
}


Document* GFFFormat::loadDocument( IOAdapter* io, TaskStateInfo& ti, const QVariantMap& _fs, DocumentLoadMode mode ){
    if( NULL == io || !io->isOpen() ) {
        ti.setError(L10N::badArgument("IO adapter"));
        return NULL;
    }
    QList<GObject*> objects;

    load(io, objects, ti);

    if (ti.hasErrors() || ti.cancelFlag) {
        qDeleteAll(objects);
        return NULL;
    }

    Document* doc = new Document( this, io->getFactory(), io->getURL(), objects);
    return doc;
}

void GFFFormat::load(IOAdapter* io, QList<GObject*>& objects, TaskStateInfo& si){
    QStringList words;
    bool isOk;
    gauto_array<char> buff = new char[READ_BUFF_SIZE];
    int len = io->readLine(buff.data, READ_BUFF_SIZE);
    buff.data[len] = '\0';
    QString qstrbuf(buff.data);
    words = parseLine(qstrbuf);
    //header validation
    if(words.size() < 2){
        si.setError(tr("Parsing error: invalid header"));
        return;
    }
    if(!words[0].startsWith("##gff-version")){
        si.setError(tr("Parsing error: file not contains version header"));
        return;
    }
    int ver = words[1].toInt(&isOk);
    if(!isOk){
        si.setError(tr("Parsing error: format version is not integer"));
        return;
    }
    //is version supported
    if(ver != 3){
        si.setError(tr("Parsing error: version %1 not supported").arg(ver));
        return;
    }

    int lineNumber = 1;
    while((len = io->readLine(buff.data, READ_BUFF_SIZE)) > 0){
        qstrbuf.clear();
        buff.data[len] = '\0';
        qstrbuf = QString(buff.data);
        //retrieving annotations from  document
        words = parseLine(qstrbuf);
        if(!words[0].startsWith("##")){
            if(words.size() != 9){
                si.setError(tr("Parsing error: too few fields at line %1").arg(lineNumber));
                return;
            }
            //annotation's region
            int start = words[3].toInt(&isOk);
            if(!isOk){
                si.setError(tr("Parsing error: start position at line %1 is not integer").arg(lineNumber));
                return;
            }

            int end = words[4].toInt(&isOk);
            if(!isOk){
                si.setError(tr("Parsing error: end position at line %1 is not integer").arg(lineNumber));
                return;
            }

            if(start > end){
                si.setError(tr("Parsing error: incorrect annotation region at line %1").arg(lineNumber));
                return;
            }

            start--;
            LRegion range(start, end - start);

            QString groupName = words[2];
            QString annName = groupName;
            //annotation's qualifiers from attributes
            SharedAnnotationData d(new AnnotationData());
            if(words[8] != "."){
                QStringList pairs = words[8].split(";", QString::SkipEmptyParts);
                bool qualName = false;
                foreach(QString p, pairs){
                    QStringList qual = p.split("=", QString::SkipEmptyParts);
                    if(qual.size() != 2){
                        si.setError(tr("Parsing error: one of the qualifiers in attributes are incorrect").arg(lineNumber));
                        return;
                    }
                    d->qualifiers.append(Qualifier(qual[0], qual[1]));
                }
            }

            d->location.append(range);
            d->name = annName;		

            QString atoName = words[0];
            AnnotationTableObject *ato = NULL;
            foreach(GObject *ob, objects){
                if(ob->getGObjectName() == atoName){
                    ato = (AnnotationTableObject *)ob;
                }
            }
            if(!ato){
                ato = new AnnotationTableObject(atoName);
                objects.append(ato);
            }

            //qualifiers from columns
            if(words[1] != "."){
                d->qualifiers.append(Qualifier("source", words[1]));
            }

            if(words[5] != "."){
                d->qualifiers.append(Qualifier("score", words[5]));
            }

            if(words[7] != "."){
                d->qualifiers.append(Qualifier("phase", words[7]));
            }

            Annotation *a = new Annotation(d);
            //strand detection
            if(words[6] == "-"){
                a->setOnComplementStrand(true);
            }

            ato->addAnnotation(a, groupName);
        }
        lineNumber++;
    }
}

bool GFFFormat::checkRawData(const QByteArray& rawData) const {
    const char* data = rawData.constData();
    int size = rawData.size();
    int n = TextUtils::skip(TextUtils::WHITES, data, size);
    int newSize = size - n;
    const char* newData = data + n;
    if (newSize <= 13) {
        return false;
    }
    QString header(QByteArray(newData, 13));
    if(header != "##gff-version"){
        return false;
    }
    bool hasBinaryData = TextUtils::contains(TextUtils::BINARY, data, size);
    return hasBinaryData ? false : true;
}

QStringList GFFFormat::parseLine( QString line ){
    QChar prev('a'); //as default value not empty char
    QString pair;
    QStringList result;
    QString word;

    foreach(QChar c, line){
        pair.clear();
        pair.append(prev);
        pair.append(c);
        if((c == '\t') || (pair == " \t") || (pair == "  ")){
            if((word != "  ") && (word != " ") && !word.isEmpty()){
                result.append(word);
            }
            word.clear();
        }else{
            word.append(c);
        }
        prev = c;
    }
    if((word != "  ") && (word != " ") && !word.isEmpty()){
        result.append(word);
    }
    return result;
}


} //namespace

