/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "MSAUtils.h"

#include <gobjects/DNASequenceObject.h>
#include <util_text/TextUtils.h>


#include <datatype/MAlignment.h>

namespace GB2 {

bool MSAUtils::equalsIgnoreGaps(const QByteArray& seq, int startPos, const QByteArray& pat) {
    int sLen = seq.size();
    int pLen = pat.size();
    for (int i = startPos, j = 0; i  < sLen && j < pLen; i++, j++) {
        char c1 = seq[i];
        char c2 = pat[j];
        while(c1 == MAlignment_GapChar && ++i < sLen) {
            c1 = seq[i];
        }
        if (c1 != c2) {
            return false;
        }
    }
    return true;
}


MAlignment MSAUtils::seq2ma( const QList<GObject*>& list, QString& err ) {
    MAlignment ma(MA_OBJECT_NAME);
    foreach(GObject* obj, list) {
        DNASequenceObject* dnaObj = qobject_cast<DNASequenceObject*>(obj);
        const DNASequence seq = dnaObj->getDNASequence();
        if (ma.alphabet == NULL) {
            ma.alphabet = dnaObj->getAlphabet();
        } else {
            ma.alphabet = DNAAlphabet::deriveCommonAlphabet(ma.alphabet, dnaObj->getAlphabet());
            if (ma.alphabet == NULL) {
                err = tr("Sequences have different alphabets.");
                break;
            }
        }
        if (seq.length() > MAX_ALI_LEN) {
            err = tr("Sequence is too large for alignment: %1").arg(seq.getName());
            break;
        }
        ma.alignedSeqs.append(MAlignmentItem(dnaObj->getGObjectName(), seq.seq));
    }

    if (err.isEmpty()) {
        ma.normalizeModel();
    } else {
        ma.clear();
    }

    return ma;
}

QList<DNASequence> MSAUtils::ma2seq(const MAlignment& ma, bool trimGaps) {
    QList<DNASequence> lst;
    QBitArray gapCharMap = TextUtils::createBitMap(MAlignment_GapChar);
    foreach(const MAlignmentItem& i, ma.alignedSeqs) {
        DNASequence s(i.name, i.sequence, ma.alphabet);
        if (trimGaps) {
            int newLen = TextUtils::remove(s.seq.data(), s.length(), gapCharMap);
            s.seq.resize(newLen);
        }
        lst << s;
    }
    return lst;
}

void MSAUtils::removeRegionFromAlignment( MAlignment& msa, int startBase, int startSeq, int width, int height )
{
    assert( (startBase >= 0) && (width <= msa.getLength()) );
    assert( (startSeq >= 0) && (height <= msa.getNumSequences()) );
    assert(msa.isNormalized());
    // do nothing if whole alignment is being removed
    if ( (msa.getNumSequences() == width) && (msa.getLength() == height) ) {
        assert(0);
        return;
    }

    int endSeq = startSeq + height - 1;
    assert(endSeq < msa.getNumSequences());

    if ( width == msa.getLength() ) {
        // special case - removing sequences
        for (int i = startSeq; i <= endSeq; ++i) {
            const MAlignmentItem& item = msa.alignedSeqs.at(startSeq);
            msa.alignedSeqs.removeAt(startSeq);

        }
    } else {
        for (int i = startSeq; i <= endSeq; ++i) {
            MAlignmentItem& item = msa.alignedSeqs[i];
            item.sequence.remove(startBase, width);
            if (height == msa.getNumSequences()) {
                // Normalization is not required
                continue;
            }   else {
                // Normalize sequence by adding gaps to its end
                QByteArray gaps(width, MAlignment_GapChar);
                item.sequence.append(gaps);
            }
        }

    }
}
}//namespace
