/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "MSAEditor.h"
#include "MSAEditorTasks.h"
#include "MSAEditorFactory.h"
#include "MSAEditorState.h"
#include "MSAEditorConsensusArea.h"
#include "MSAEditorOffsetsView.h"
#include "MSAEditorSequenceArea.h"
#include "MSAEditorNameList.h"
#include "MSAEditorStatusBar.h"
#include "MSAEditorUndoFramework.h"
#include "MSAEditorTreeViewer.h"

#include <gobjects/MAlignmentObject.h>
#include <util_gui/GUIUtils.h>
#include <util_gui/DialogUtils.h>
#include <core_api/AppContext.h>
#include <gobjects/PhyTreeObject.h>

#include <QtCore/QEvent>

#include <QtGui/QLabel>
#include <QtGui/QPainter>
#include <QtGui/QVBoxLayout>
#include <QtGui/QHBoxLayout>
#include <QtGui/QGridLayout>
#include <QtGui/QResizeEvent>
#include <QtGui/QToolBar>
#include <QtGui/QFileDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QFontDialog>

#include <core_api/IOAdapter.h>
#include <core_api/Counter.h>
#include <core_api/DocumentModel.h>
#include <core_api/DocumentFormats.h>
#include <core_api/ProjectModel.h>
#include <core_api/Settings.h>
#include <util_gui/GScrollBar.h>

namespace GB2 {

#define MIN_FONT_SIZE 6
#define MAX_FONT_SIZE 24

#define SETTINGS_ROOT QString("msaeditor/")
#define SETTINGS_FONT_FAMILY    "font_family"
#define SETTINGS_FONT_SIZE      "font_size"
#define SETTINGS_FONT_ITALIC    "font_italic"
#define SETTINGS_FONT_BOLD      "font_bold"

#define DEFAULT_FONT_FAMILY "Verdana" 
#define DEFAULT_FONT_SIZE 10

/* TRANSLATOR GB2::MSAEditor */ 

const char* MSAEditor::TreeRelation = "Tree";
const char* MSAEditor::MSARelation = "MSA";
const char* MSAEditor::ActiveTreeRelation = "ActiveTree";
const char* MSAEditor::ActiveMSARelation = "ActiveMSA";

MSAEditor::MSAEditor(const QString& viewName, MAlignmentObject* msaObj, PhyTreeObject* phyObj)
: GObjectView(MSAEditorFactory::ID, viewName), ui(NULL), msaObject(msaObj), phyObject(phyObj) {
    init();
}

MSAEditor::MSAEditor(const QString& viewName, GObject* obj)
: GObjectView(MSAEditorFactory::ID, viewName), ui(NULL) {
    
    if (obj->getGObjectType() == GObjectTypes::MULTIPLE_ALIGNMENT) {
        msaObject = qobject_cast<MAlignmentObject*>(obj);
        phyObject = NULL;
    }
    if (obj->getGObjectType() == GObjectTypes::PHYLOGENETIC_TREE) {
        phyObject = qobject_cast<PhyTreeObject*>(obj);
        msaObject = NULL;
    }
    init();
    GCOUNTER(cvar,tvar,"MSAEditor");
}

void MSAEditor::init() {
    if (msaObject == NULL) {
        QList<GObjectRelation> rel = phyObject->findRelatedObjectsByRole(MSAEditor::ActiveMSARelation);
        if (!rel.isEmpty()) {
            GObjectReference ref = rel.first().ref;
            assert(ref.objType == GObjectTypes::MULTIPLE_ALIGNMENT);
            Document *d = AppContext::getProject()->findDocumentByURL(ref.docUrl);
            if (d != NULL) {
                msaObject = qobject_cast<MAlignmentObject*>(d->findGObjectByName(ref.objName));
            }
            if (msaObject == NULL) {
                phyObject->removeObjectRelation(rel.first());
            }
        }
    }
    if (phyObject == NULL) {
        QList<GObjectRelation> rel = msaObject->findRelatedObjectsByRole(MSAEditor::ActiveTreeRelation);
        if (!rel.isEmpty()) {
            assert(rel.size() == 1);
            GObjectReference ref = rel.first().ref;
            assert(ref.objType == GObjectTypes::PHYLOGENETIC_TREE);
            Document *d = AppContext::getProject()->findDocumentByURL(ref.docUrl);
            if (d != NULL) {
                phyObject = qobject_cast<PhyTreeObject*>(d->findGObjectByName(ref.objName));
            }
            if (phyObject == NULL) {
                msaObject->removeObjectRelation(rel.first());
            }
        }
    }

    if (msaObject) {
        objects.append(msaObject);
        requiredObjects.append(msaObject);
    }
    if (phyObject) {
        objects.append(phyObject);
        requiredObjects.append(phyObject);
    }
    connect(AppContext::getProject(), SIGNAL(si_documentAdded(Document*)), SLOT(sl_documentAdded(Document*)));

    increaseFontAction = new QAction(QIcon(":core/images/zoom_in.png"), tr("Increase Font"), this);
    connect(increaseFontAction, SIGNAL(triggered()), SLOT(sl_zoomIn()));

    decreaseFontAction = new QAction(QIcon(":core/images/zoom_out.png"), tr("Decrease Font"), this);
    connect(decreaseFontAction, SIGNAL(triggered()), SLOT(sl_zoomOut()));

    changeFontAction = new QAction(QIcon(":core/images/font.png"), tr("Change Font"), this);
    connect(changeFontAction, SIGNAL(triggered()), SLOT(sl_changeFont()));

    resetFontAction = new QAction(QIcon(":core/images/font_default.png"), tr("Reset Font"), this);
    connect(resetFontAction, SIGNAL(triggered()), SLOT(sl_resetFont()));

    Settings* s = AppContext::getSettings();
    font.setFamily(s->getValue(SETTINGS_ROOT + SETTINGS_FONT_FAMILY, DEFAULT_FONT_FAMILY).toString());
    font.setPointSize(s->getValue(SETTINGS_ROOT + SETTINGS_FONT_SIZE, DEFAULT_FONT_SIZE).toInt());
    font.setItalic(s->getValue(SETTINGS_ROOT + SETTINGS_FONT_ITALIC, false).toBool());
    font.setBold(s->getValue(SETTINGS_ROOT + SETTINGS_FONT_BOLD, false).toBool());
    increaseFontAction->setEnabled(font.pointSize() < MAX_FONT_SIZE);
    decreaseFontAction->setEnabled(font.pointSize() > MIN_FONT_SIZE);
}

int MSAEditor::getRowHeight() const {
    QFontMetrics fm(font);
    return fm.height() * 5 / 4;
}

void MSAEditor::increaseFont() {
    increaseFontAction->trigger();
}

void MSAEditor::decreaseFont() {
    decreaseFontAction->trigger();
}

int MSAEditor::getColumnWidth() const {
    QFontMetrics fm(font);
    return fm.width('W') * 5 / 4;
}

void MSAEditor::sl_zoomIn() {
    int pSize = font.pointSize();
    if (pSize < MAX_FONT_SIZE) {
        font.setPointSize(pSize+1);
        setFont(font);
    }
}

void MSAEditor::sl_zoomOut() {
    int pSize = font.pointSize();
    if (pSize > MIN_FONT_SIZE) {
        font.setPointSize(pSize-1);
        setFont(font);
    }
}

static void saveFont(const QFont& f) {
    Settings* s = AppContext::getSettings();
    s->setValue(SETTINGS_ROOT + SETTINGS_FONT_FAMILY, f.family());
    s->setValue(SETTINGS_ROOT + SETTINGS_FONT_SIZE, f.pointSize());
    s->setValue(SETTINGS_ROOT + SETTINGS_FONT_ITALIC, f.italic());
    s->setValue(SETTINGS_ROOT + SETTINGS_FONT_BOLD, f.bold());
}

void MSAEditor::setFont(const QFont& f) {
    int pSize = f.pointSize();
    font = f;
    font.setPointSize(qBound(MIN_FONT_SIZE, pSize, MAX_FONT_SIZE));
    increaseFontAction->setEnabled(pSize < MAX_FONT_SIZE);
    decreaseFontAction->setEnabled(pSize > MIN_FONT_SIZE);
    emit si_fontChanged(f);
    saveFont(font);
}

void MSAEditor::sl_changeFont() {
    bool ok = false;
    QFont f = QFontDialog::getFont(&ok, font, widget, tr("Select font for alignment"));
    if (!ok) {
        return;
    }
    setFont(f);
}

void MSAEditor::sl_resetFont() {
    QFont f(DEFAULT_FONT_FAMILY, DEFAULT_FONT_SIZE);
    setFont(f);
}

void MSAEditor::sl_documentAdded(Document *d) {
    connect(d, SIGNAL(si_objectAdded(GObject*)), SLOT(sl_objectAdded(GObject*)));
}

void MSAEditor::sl_objectAdded(GObject *o) {
    if (canAddObject(o) && !ui->treeViewer->isCreated()) {
        addObject(o);
    }
}

MSAEditor::~MSAEditor() {
}

void MSAEditor::buildStaticToolbar(QToolBar* tb) {
    tb->addAction(increaseFontAction);
    tb->addAction(decreaseFontAction);
    tb->addAction(changeFontAction);
    tb->addAction(resetFontAction);
    GObjectView::buildStaticToolbar(tb);
}

void MSAEditor::buildStaticMenu(QMenu* m) {
    addLoadMenu(m);
    addCopyMenu(m);
    addEditMenu(m);
    addAlignMenu(m);
    addStatisticsMenu(m);
    addViewMenu(m);
    addAdvancedMenu(m);

    GObjectView::buildStaticMenu(m);

    GUIUtils::disableEmptySubmenus(m);
}


void MSAEditor::addCopyMenu(QMenu* m) {
    QMenu* cm = m->addMenu(tr("copy_menu"));
    cm->menuAction()->setObjectName(MSAE_MENU_COPY);
}

void MSAEditor::addEditMenu(QMenu* m) {
    QMenu* em = m->addMenu(tr("edit_menu"));
    em->menuAction()->setObjectName(MSAE_MENU_EDIT);
}

void MSAEditor::addViewMenu(QMenu* m) {
    QMenu* em = m->addMenu(tr("View"));
    em->menuAction()->setObjectName(MSAE_MENU_VIEW);
    if (ui->offsetsView)
        em->addAction(ui->offsetsView->getToggleColumnsViewAction());
}

void MSAEditor::addAlignMenu(QMenu* m) {
    QMenu* em = m->addMenu(tr("align_menu"));
    em->setIcon(QIcon(":core/images/align.png"));
    em->menuAction()->setObjectName(MSAE_MENU_ALIGN);
}

void MSAEditor::addAdvancedMenu(QMenu* m) {
    QMenu* em = m->addMenu(tr("advanced_menu"));
    em->menuAction()->setObjectName(MSAE_MENU_ADVANCED);
}

void MSAEditor::addStatisticsMenu(QMenu* m) {
    QMenu* em = m->addMenu(tr("Statistics"));
    em->setIcon(QIcon(":core/images/chart_bar.png"));
    em->menuAction()->setObjectName(MSAE_MENU_STATISTICS);
}

void MSAEditor::addLoadMenu( QMenu* m )
{
    QMenu* lsm = m->addMenu(tr("Add"));
    lsm->menuAction()->setObjectName(MSAE_MENU_LOAD);
}



Task* MSAEditor::updateViewTask(const QString& stateName, const QVariantMap& stateData) {
    return new UpdateMSAEditorTask(this, stateName, stateData);
}

QVariantMap MSAEditor::saveState() {
    return MSAEditorState::saveState(this);
}

QWidget* MSAEditor::createWidget() {
    assert(ui == NULL);
    ui = new MSAEditorUI(this);
    connect(ui , SIGNAL(customContextMenuRequested(const QPoint &)), SLOT(sl_onContextMenuRequested(const QPoint &)));
    ui->installEventFilter(this);
    ui->setAcceptDrops(true);
    return ui;
}

int MSAEditor::getAlignmentLen() const {
    return msaObject->getMAlignment().getLength();
}

int MSAEditor::getNumSequences() const {
    return msaObject->getMAlignment().getNumSequences();
}

void MSAEditor::sl_onContextMenuRequested(const QPoint & pos) {
    Q_UNUSED(pos);

    QMenu m;

    addLoadMenu(&m);
    addCopyMenu(&m);
    addEditMenu(&m);
    addAlignMenu(&m);
    addStatisticsMenu(&m);
    addViewMenu(&m);
    addAdvancedMenu(&m);

    emit si_buildPopupMenu(this, &m);

    GUIUtils::disableEmptySubmenus(&m);

    m.exec(QCursor::pos());
}

bool MSAEditor::canAddObject(GObject* obj) {
    if (obj->getGObjectType() == GObjectTypes::PHYLOGENETIC_TREE) {
        if (msaObject == NULL)
            return false;
        if (msaObject->hasObjectRelation(obj, MSAEditor::TreeRelation))
            return true;
    }
    return false;
}

bool MSAEditor::eventFilter(QObject* o, QEvent* e) {
    if (o == ui) {
        if (e->type() == QEvent::DragEnter || e->type() == QEvent::Drop) {
            QDropEvent* de = (QDropEvent*)e;
            const QMimeData* md = de->mimeData();
            const GObjectMimeData* gomd = qobject_cast<const GObjectMimeData*>(md);
            if (gomd != NULL) {
                if (e->type() == QEvent::DragEnter) {
                    de->acceptProposedAction();
                } else {
                    GObject* obj = gomd->objPtr.data();
                    if (obj!=NULL) {
                        QString err = addObject(obj);
                        if (!err.isEmpty()) {
                            QMessageBox::critical(NULL, tr("error"), err);
                        } 
                    }
                }
            }
        } 
    }

    return false;
}

QString MSAEditor::addObject(GObject* o) {
    if (o->getGObjectType() == GObjectTypes::UNLOADED) {
        AppContext::getTaskScheduler()->registerTopLevelTask(new AddToViewTask(this, o));
        return "";
    }
    if (o->getGObjectType() == GObjectTypes::PHYLOGENETIC_TREE) {
        PhyTreeObject* obj = qobject_cast<PhyTreeObject*>(o);
        if (obj != NULL && obj != phyObject && msaObject->hasObjectRelation(obj, MSAEditor::TreeRelation)) {
            setTree(obj);
            ui->treeViewer->setTree(obj);
        }
    }
    return "";
}

void MSAEditor::setTree(PhyTreeObject* obj) {
    if (msaObject != NULL && phyObject != NULL) {
//        phyObject->removeObjectRelation(GObjectRelation(msaObject, MSAEditor::ActiveMSARelation));
        msaObject->removeObjectRelation(GObjectRelation(phyObject, MSAEditor::ActiveTreeRelation));
    }
    if (phyObject != NULL) {
        objects.removeOne(phyObject);
        requiredObjects.removeOne(phyObject);
    }
    phyObject = obj;
    objects.append(phyObject);
    requiredObjects.append(phyObject);
    if (msaObject != NULL) {
        if (!msaObject->hasObjectRelation(phyObject, MSAEditor::TreeRelation)) {
            msaObject->addObjectRelation(phyObject, MSAEditor::TreeRelation);
        }
        if (!phyObject->hasObjectRelation(msaObject, MSAEditor::MSARelation)) {
            phyObject->addObjectRelation(msaObject, MSAEditor::MSARelation);
        }
        msaObject->addObjectRelation(phyObject, MSAEditor::ActiveTreeRelation);
        phyObject->addObjectRelation(msaObject, MSAEditor::ActiveMSARelation);
    }
}

const QRect MSAEditor::getCurrentSelection() const
{
    return ui->seqArea->getSelection().getRect();
}



//////////////////////////////////////////////////////////////////////////
MSAEditorUI::MSAEditorUI(MSAEditor* _editor): editor(_editor), seqArea(NULL), offsetsView(NULL), statusWidget(NULL) {
    undoFWK = new MSAEditorUndoFramework(this, editor->getMSAObject());

    setContextMenuPolicy(Qt::CustomContextMenu);
    setMinimumSize(300, 200);

    GScrollBar* cvBar = new GScrollBar(Qt::Vertical);

    QWidget *seqAreaContainer = NULL, *nameAreaContainer = NULL, *label = NULL;
    if (editor->getMSAObject()) {
        GScrollBar* shBar = new GScrollBar(Qt::Horizontal);
        QScrollBar* nhBar = new QScrollBar(Qt::Horizontal);
        seqArea = new MSAEditorSequenceArea(this, shBar, cvBar);
        nameList = new MSAEditorNameList(this, nhBar);
        consArea = new MSAEditorConsensusArea(this);
        offsetsView = new MSAEditorOffsetsViewController(this, editor, seqArea);
        statusWidget = new MSAEditorStatusWidget(editor->getMSAObject(), seqArea);

        seqArea->installEventFilter(this);
        nameList->installEventFilter(this);
        offsetsView->getLeftWidget()->installEventFilter(this);
        offsetsView->getRightWidget()->installEventFilter(this);

        QWidget* label1 = createLabelWidget();
        QWidget* label2 = createLabelWidget();
        label1->setMinimumHeight(consArea->height());
        label2->setMinimumHeight(consArea->height());
        offsetsView->getLeftWidget()->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::MinimumExpanding);
        offsetsView->getRightWidget()->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::MinimumExpanding);

        QGridLayout* seqAreaLayout = new QGridLayout();
        seqAreaLayout->setMargin(0);
        seqAreaLayout->setSpacing(0);
        seqAreaLayout->addWidget(consArea, 0, 1);
        seqAreaLayout->addWidget(label1, 0, 0);
        seqAreaLayout->addWidget(offsetsView->getLeftWidget(), 1, 0, 2, 1);
        seqAreaLayout->addWidget(seqArea, 1, 1);
        seqAreaLayout->addWidget(label2, 0, 2);
        seqAreaLayout->addWidget(offsetsView->getRightWidget(), 1, 2, 2, 1);
        seqAreaLayout->addWidget(shBar, 2, 1, 1, 1);
        seqAreaContainer = new QWidget();
        seqAreaContainer->setLayout(seqAreaLayout);

        label = createLabelWidget(tr("Consensus"));
        label->setMinimumHeight(consArea->height());
        nameList->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::MinimumExpanding);

        QVBoxLayout* nameAreaLayout = new QVBoxLayout();
        nameAreaLayout->setMargin(0);
        nameAreaLayout->setSpacing(0);
        nameAreaLayout->addWidget(label);
        nameAreaLayout->addWidget(nameList);
        nameAreaLayout->addWidget(nhBar);
        nameAreaContainer = new QWidget();
        nameAreaContainer->setLayout(nameAreaLayout);

        label = createLabelWidget(tr("Tree view"));
        label->setMinimumHeight(consArea->height());
        statusWidget->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);
    }

    treeViewer = new MSAEditorTreeViewer(this, cvBar);
    treeViewer->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::MinimumExpanding);

    QVBoxLayout* treeViewerLayout = new QVBoxLayout();
    treeViewerLayout->setMargin(0);
    treeViewerLayout->setSpacing(0);
    if (label)
        treeViewerLayout->addWidget(label);
    treeViewerLayout->addWidget(treeViewer);
    QWidget *treeViewerContainer = new QWidget();
    treeViewerContainer->setLayout(treeViewerLayout);
    if (editor->getPhyObject() == NULL)
        treeViewerContainer->hide();

    splitter = new QSplitter(Qt::Horizontal);
    splitter->addWidget(treeViewerContainer);
    if (editor->getMSAObject()) {
        splitter->addWidget(nameAreaContainer);
        splitter->addWidget(seqAreaContainer);
        splitter->setStretchFactor(0, 0);
        splitter->setStretchFactor(1, 0);
        splitter->setStretchFactor(2, 1);
        splitter->setSizes(QList<int>()<<MSAEditorTreeViewer::INITIAL_WIDTH<<220<<220); //todo: make initial namelist size depends on font
    } else {
        splitter->setSizes(QList<int>()<<MSAEditorTreeViewer::INITIAL_WIDTH);
    }
    splitter->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);

    if (editor->getMSAObject()) {
        label = createLabelWidget();
        label->setMinimumHeight(consArea->height());
        cvBar->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::MinimumExpanding);
    }

    QGridLayout *mainLayout = new QGridLayout();
    mainLayout->setMargin(0);
    mainLayout->setSpacing(0);
    if (editor->getMSAObject()) {
        mainLayout->addWidget(splitter, 0, 0, 2, 1);
        mainLayout->addWidget(cvBar, 1, 1);
        mainLayout->addWidget(label, 0, 1);
        mainLayout->addWidget(statusWidget, 2, 0, 1, 2);
    } else {
        mainLayout->addWidget(splitter, 0, 0);
//        mainLayout->addWidget(cvBar, 0, 1);
    }
    QWidget *mainWidget = new QWidget();
    mainWidget->setLayout(mainLayout);
    setLayout(mainLayout);
}

QWidget* MSAEditorUI::createLabelWidget(const QString& text, Qt::Alignment ali) const {
    return new MSALabelWidget(this, text, ali);
}

QAction* MSAEditorUI::getUndoAction() const {
    return undoFWK->getUndoAction();
}

QAction* MSAEditorUI::getRedoAction() const {
    return undoFWK->getRedoAction();
}

void MSAEditorUI::saveScreenshot(QString filename, QString format) {
    QRect screenRect = nameList->geometry();
    screenRect.setBottomRight(seqArea->parentWidget()->geometry().bottomRight());
    QPixmap curPixMap = QPixmap::grabWidget(this, screenRect);
    curPixMap.save(filename, format.toAscii().constData());
}

MSALabelWidget::MSALabelWidget(const MSAEditorUI* _ui, const QString & _t, Qt::Alignment _a) 
: ui(_ui), text(_t), ali(_a)
{
    connect(ui->getEditor(), SIGNAL(si_fontChanged(const QFont&)), SLOT(sl_fontChanged()));
}

void MSALabelWidget::paintEvent(QPaintEvent *) {
    QPainter p(this);
    p.fillRect(rect(), Qt::white);
    if (!text.isEmpty()) {
        p.setFont(ui->getEditor()->getFont());
        p.drawText(rect(), text, ali);
    }
}

void MSALabelWidget::sl_fontChanged() {
    update();
    setMinimumHeight(ui->consArea->height());
}

}//namespace

