/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#ifndef _GB2_MSA_EDITOR_H_
#define _GB2_MSA_EDITOR_H_

#include <core_api/ObjectViewModel.h>
#include <core_api/LRegion.h>
#include <datatype/PhyTree.h>

#include <QtGui/QMenu>
#include <QtGui/QSplitter>

namespace GB2 {

class MAlignmentObject;
class PhyTreeObject;
class MSAEditorUI;
class MSAEditorSequenceArea;
class MSAEditorConsensusArea;
class MSAEditorNameList;
class MSAEditorOffsetsViewController;
class MSAEditorStatusWidget;
class MSAEditorUndoFramework;
class MSAEditorTreeViewer;

#define MSAE_MENU_COPY          "MSAE_MENU_COPY"
#define MSAE_MENU_EDIT          "MSAE_MENU_EDIT"
#define MSAE_MENU_VIEW          "MSAE_MENU_VIEW"
#define MSAE_MENU_ALIGN         "MSAE_MENU_ALIGN"
#define MSAE_MENU_STATISTICS    "MSAE_MENU_STATISTICS"
#define MSAE_MENU_ADVANCED      "MSAE_MENU_ADVANCED"
#define MSAE_MENU_LOAD          "MSAE_MENU_LOAD_SEQ"

class GB2_COREAPI_EXPORT MSAEditor : public GObjectView {
    Q_OBJECT
public:
    MSAEditor(const QString& viewName, GObject* obj);
    MSAEditor(const QString& viewName, MAlignmentObject* msaObj, PhyTreeObject* phyObj);
    ~MSAEditor();

    virtual void buildStaticToolbar(QToolBar* tb);

    virtual void buildStaticMenu(QMenu* m);

    virtual Task* updateViewTask(const QString& stateName, const QVariantMap& stateData);

    virtual QVariantMap saveState();
    
    // returns active in the editor MSA object or NULL
	MAlignmentObject* getMSAObject() const {return msaObject;}
    
    // returns active in the editor Phy tree object or NULL
	PhyTreeObject* getPhyObject() const {return phyObject;}

    const MSAEditorUI* getUI() const {return ui;}

    int getAlignmentLen() const;
    
    int getNumSequences() const;

    const QRect getCurrentSelection() const;
    
    virtual bool canAddObject(GObject* obj);
    
    virtual QString addObject(GObject* o);

    const QFont& getFont() {return font;}

    int getRowHeight() const;

    int getColumnWidth() const;

    void setTree(PhyTreeObject* obj);

    void increaseFont();

    void decreaseFont();

    static const char* TreeRelation;
    static const char* MSARelation;
    static const char* ActiveTreeRelation;
    static const char* ActiveMSARelation;

signals:
    void si_fontChanged(const QFont& f);

protected slots:
    void sl_onContextMenuRequested(const QPoint & pos);
    void sl_documentAdded(Document *d);
    void sl_objectAdded(GObject *o);
    void sl_zoomIn(); 
    void sl_zoomOut(); 
    void sl_changeFont(); 
    void sl_resetFont(); 
   

protected:
    virtual QWidget* createWidget();
    virtual bool eventFilter(QObject*, QEvent*);

private:
    void init();
    void addCopyMenu(QMenu* m);
    void addEditMenu(QMenu* m);
    void addViewMenu(QMenu* m);
    void addAlignMenu(QMenu* m);
    void addAdvancedMenu(QMenu* m);
    void addStatisticsMenu(QMenu* m);
    void addLoadMenu(QMenu* m);
    void setFont(const QFont& f);

    MAlignmentObject* msaObject;
    PhyTreeObject*    phyObject;
    MSAEditorUI*      ui;
    QFont             font;

    QAction*          increaseFontAction;
    QAction*          decreaseFontAction;
    QAction*          changeFontAction;
    QAction*          resetFontAction;

};

class MSAEditorUI : public QWidget {
    Q_OBJECT
    //todo: make public accessors:
    friend class MSALabelWidget;
    friend class MSAEditorSequenceArea;
    friend class MSAEditorConsensusArea;
    friend class MSAEditorNameList;
    friend class MSAEditorTreeViewer;
    friend class MSAEditor;

public:
    MSAEditorUI(MSAEditor* editor);

    QWidget* createLabelWidget(const QString& text = QString(), Qt::Alignment ali = Qt::AlignCenter) const;

    MSAEditor* getEditor() const {return editor;}

    QAction* getUndoAction() const;
    QAction* getRedoAction() const;
   
    void saveScreenshot(QString filename, QString format);

private:
    MSAEditor*                      editor;
    MSAEditorNameList*              nameList;
    MSAEditorSequenceArea*          seqArea;
    MSAEditorConsensusArea*         consArea;
    MSAEditorOffsetsViewController* offsetsView;
    MSAEditorStatusWidget*          statusWidget;
    MSAEditorTreeViewer*            treeViewer;
    QSplitter*                      splitter;

    QList<QWidget*>                 nameAreaWidgets;
    QList<QWidget*>                 lw1Widgets;
    QList<QWidget*>                 seqAreaWidgets;
    QList<QWidget*>                 lw2Widgets;
    QList<QWidget*>                 treeAreaWidgets;
    MSAEditorUndoFramework*         undoFWK;

};



class MSALabelWidget : public QWidget {
    Q_OBJECT
public:
    MSALabelWidget(const MSAEditorUI* _ui, const QString & _t, Qt::Alignment _a);
    
    const MSAEditorUI*  ui;
    QString             text;
    Qt::Alignment       ali;

protected slots:
    void sl_fontChanged();

protected:
    void paintEvent(QPaintEvent *e);
};


}//namespace;

#endif
