/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "MSAEditorFactory.h"
#include "MSAEditor.h"
#include "MSAEditorTasks.h"
#include "MSAEditorState.h"

#include <core_api/Log.h>
#include <core_api/DocumentModel.h>
#include <core_api/AppContext.h>
#include <core_api/ProjectModel.h>

#include <gobjects/MAlignmentObject.h>
#include <gobjects/PhyTreeObject.h>
#include <gobjects/UnloadedObject.h>

#include <selection/SelectionUtils.h>

namespace GB2 {

/* TRANSLATOR GB2::MSAEditor */
/* TRANSLATOR GB2::ObjectViewTask */

static LogCategory log(ULOG_CAT_MSA);

const GObjectViewFactoryId MSAEditorFactory::ID("MSAEditor");

MSAEditorFactory::MSAEditorFactory() 
: GObjectViewFactory(ID, tr("Alignment editor"))
{
}

bool MSAEditorFactory::canCreateView(const MultiGSelection& multiSelection) {
	bool hasMSADocuments = !SelectionUtils::findDocumentsWithObjects(
                                GObjectTypes::MULTIPLE_ALIGNMENT, &multiSelection, UOF_LoadedAndUnloaded, true).isEmpty();
    if (hasMSADocuments) {
        return true;
    }
	hasMSADocuments = !SelectionUtils::findDocumentsWithObjects(
                                GObjectTypes::PHYLOGENETIC_TREE, &multiSelection, UOF_LoadedAndUnloaded, true).isEmpty();
    if (hasMSADocuments) {
        return true;
    }
    return false;
}

Task* MSAEditorFactory::createViewTask(const MultiGSelection& multiSelection, bool single) {
	QSet<GObject*> msaObjects = SelectionUtils::findObjects(GObjectTypes::MULTIPLE_ALIGNMENT, &multiSelection, UOF_LoadedAndUnloaded);
    QSet<Document*> docsWithMSA = SelectionUtils::findDocumentsWithObjects(GObjectTypes::MULTIPLE_ALIGNMENT, 
                                                            &multiSelection, UOF_LoadedAndUnloaded, false);
	QSet<GObject*> phyObjects = SelectionUtils::findObjects(GObjectTypes::PHYLOGENETIC_TREE, &multiSelection, UOF_LoadedAndUnloaded);
    QSet<Document*> docsWithPhy = SelectionUtils::findDocumentsWithObjects(GObjectTypes::PHYLOGENETIC_TREE, 
                                                            &multiSelection, UOF_LoadedAndUnloaded, false);

    foreach(Document* doc, docsWithMSA) {
        QList<GObject*> docObjs = doc->findGObjectByType(GObjectTypes::MULTIPLE_ALIGNMENT, UOF_LoadedAndUnloaded);
        assert(!docObjs.isEmpty());
        msaObjects += docObjs.toSet();
    }

    foreach(Document* doc, docsWithPhy) {
        QList<GObject*> docObjs = doc->findGObjectByType(GObjectTypes::PHYLOGENETIC_TREE, UOF_LoadedAndUnloaded);
        assert(!docObjs.isEmpty());
        phyObjects += docObjs.toSet();
    }

    return new OpenMSAEditorsTask(msaObjects, phyObjects);
}

bool MSAEditorFactory::isStateInSelection(const MultiGSelection& multiSelection, const QVariantMap& stateData) {
	MSAEditorState state(stateData);
	if (!state.isValid()) {
		return false;
	}
	GObjectReference msaRef = state.getMSAObject();
	GObjectReference phyRef = state.getPhyObject();
	Document* msaDoc = msaRef.isValid() ? AppContext::getProject()->findDocumentByURL(msaRef.docUrl) : NULL;
	Document* phyDoc = phyRef.isValid() ? AppContext::getProject()->findDocumentByURL(phyRef.docUrl) : NULL;
	if (msaDoc == NULL && phyDoc == NULL) { //todo: accept to use invalid state removal routines of ObjectViewTask ???
		return false;
	}
	//check that document is in selection
	QList<Document*> selectedDocs = SelectionUtils::getSelectedDocs(multiSelection);
	if ((msaDoc == NULL || selectedDocs.contains(msaDoc)) && (phyDoc == NULL || selectedDocs.contains(phyDoc))) {
		return true;
	}
	//check that object is in selection
	QList<GObject*> selectedObjects = SelectionUtils::getSelectedObjects(multiSelection);
	GObject* msaObj = msaRef.isValid() ? msaDoc->findGObjectByName(msaRef.objName) : NULL;
	GObject* phyObj = phyRef.isValid() ? phyDoc->findGObjectByName(phyRef.objName) : NULL;
	return msaObj != NULL && selectedObjects.contains(msaObj) && phyObj != NULL && selectedObjects.contains(phyObj);
}

Task* MSAEditorFactory::createViewTask(const QString& viewName, const QVariantMap& stateData) {
	return new OpenSavedMSAEditorTask(viewName, stateData);
}

}//namespace

