/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "MSAEditorTasks.h"
#include "MSAEditor.h"
#include "MSAEditorFactory.h"
#include "MSAEditorState.h"

#include <core_api/Log.h>
#include <core_api/AppContext.h>
#include <core_api/ProjectModel.h>
#include <core_api/DocumentModel.h>

#include <gobjects/MAlignmentObject.h>
#include <gobjects/PhyTreeObject.h>
#include <gobjects/UnloadedObject.h>
#include <gobjects/GObjectTypes.h>
#include <util_tasks/LoadDocumentTask.h>

#include <QtCore/QSet>

namespace GB2 {

/* TRANSLATOR GB2::MSAEditor */
/* TRANSLATOR GB2::ObjectViewTask */

static LogCategory log(ULOG_CAT_MSA);

//////////////////////////////////////////////////////////////////////////
/// open new view

OpenMSAEditorsTask::OpenMSAEditorsTask(QSet<GObject*> msaObjs, QSet<GObject*> phyObjs) 
: Task(tr("Load MSA editors"), TaskFlag_NoRun), msaObjects(msaObjs), phyObjects(phyObjs) {}

#define MAX_VIEWS 10

void OpenMSAEditorsTask::prepare() {
    for (QSet<GObject*>::iterator it = msaObjects.begin(), end = msaObjects.end(); it != end; ) {
        if ((*it)->getGObjectType() == GObjectTypes::UNLOADED) {
            Document *doc = (*it)->getDocument();
            if (!loadedDocuments.contains(doc)) {
                addSubTask(new LoadUnloadedDocumentTask(doc));
                loadedDocuments.insert(doc);
                it = msaObjects.erase(it);
                continue;
            }
        }
        ++it;
    }
    for (QSet<GObject*>::iterator it = phyObjects.begin(), end = phyObjects.end(); it != end; ) {
        if ((*it)->getGObjectType() == GObjectTypes::UNLOADED) {
            Document *doc = (*it)->getDocument();
            if (!loadedDocuments.contains(doc)) {
                addSubTask(new LoadUnloadedDocumentTask(doc));
                loadedDocuments.insert(doc);
                it = phyObjects.erase(it);
                continue;
            }
        }
        ++it;
    }
}

Task::ReportResult OpenMSAEditorsTask::report() {
	if (stateInfo.hasErrors()) {
		return ReportResult_Finished;
	}

    QSet<GObject*> allMsaObjects, allPhyObjects;
    QSet<MSAEditor*> msaEditors, oldMsaEditors;
    MWMDIManager* mdiManager = 	AppContext::getMainWindow()->getMDIManager();
    const QList<Document*> &docs = AppContext::getProject()->getDocuments();

    QList<MWMDIWindow*> wins = mdiManager->getWindows();
    foreach(MWMDIWindow* win, wins) {
        GObjectViewWindow *gwin = qobject_cast<GObjectViewWindow*>(win);
        if (gwin != NULL) {
            MSAEditor *ed = qobject_cast<MSAEditor*>(gwin->getObjectView());
            if (ed != NULL) {
                oldMsaEditors += ed;
            }
        }
    }

    foreach(Document* doc, docs) {
        allMsaObjects += doc->findGObjectByType(GObjectTypes::MULTIPLE_ALIGNMENT, UOF_LoadedOnly).toSet();
    }

    foreach(Document* doc, docs) {
        allPhyObjects += doc->findGObjectByType(GObjectTypes::PHYLOGENETIC_TREE, UOF_LoadedOnly).toSet();
    }

    foreach(Document *doc, loadedDocuments) {
        msaObjects += doc->findGObjectByType(GObjectTypes::MULTIPLE_ALIGNMENT, UOF_LoadedOnly).toSet();
        phyObjects += doc->findGObjectByType(GObjectTypes::PHYLOGENETIC_TREE, UOF_LoadedOnly).toSet();
    }

    QSet<PhyTreeObject*> phyObjs;
    foreach(GObject* o, phyObjects) {
        if (o->getGObjectType() == GObjectTypes::UNLOADED) {
            GObject* obj = o->getDocument()->findGObjectByName(o->getGObjectName());
            if (obj == NULL || obj->getGObjectType() != GObjectTypes::PHYLOGENETIC_TREE)
                continue;
            phyObjs.insert(qobject_cast<PhyTreeObject*>(obj));
            allPhyObjects.insert(obj);
        }
        if (o->getGObjectType() == GObjectTypes::PHYLOGENETIC_TREE) {
            phyObjs.insert(qobject_cast<PhyTreeObject*>(o));
            allPhyObjects.insert(o);
        }
    }

    QSet<MAlignmentObject*> msaObjs;
    foreach(GObject* o, msaObjects) {
        MAlignmentObject *mo;
        if (o->getGObjectType() == GObjectTypes::UNLOADED) {
            GObject* obj = o->getDocument()->findGObjectByName(o->getGObjectName());
            if (obj == NULL || obj->getGObjectType() != GObjectTypes::MULTIPLE_ALIGNMENT)
                continue;
            mo = qobject_cast<MAlignmentObject*>(obj);
            msaObjs.insert(mo);
        } else if (o->getGObjectType() == GObjectTypes::MULTIPLE_ALIGNMENT) {
            mo = qobject_cast<MAlignmentObject*>(o);
            msaObjs.insert(mo);
        } else {
            continue;
        }

        QSet<QString> names;
        for (int i = 0, s = mo->getMAlignment().getNumSequences(); i < s; ++i) {
            QString str = mo->getMAlignment().alignedSeqs[i].name;
            names.insert(str.replace('_', ' '));
        }

        foreach(GObject* obj, allPhyObjects) {
            PhyTreeObject *phyObj = qobject_cast<PhyTreeObject*>(obj);
            if (PhyTreeUtils::getNumSeqsFromNode(phyObj->getTree()->rootNode, names) == mo->getMAlignment().getNumSequences()) {
                phyObj->addObjectRelation(mo, MSAEditor::MSARelation);
                mo->addObjectRelation(phyObj, MSAEditor::TreeRelation);
            } else {
                phyObj->removeObjectRelation(GObjectRelation(mo, MSAEditor::MSARelation));
                mo->removeObjectRelation(GObjectRelation(phyObj, MSAEditor::TreeRelation));
            }
        }
    }

    if (!phyObjs.isEmpty()) {
        foreach(GObject* obj, allMsaObjects) {
            MAlignmentObject *mo = qobject_cast<MAlignmentObject*>(obj);
            QSet<QString> names;
            for (int i = 0, s = mo->getMAlignment().getNumSequences(); i < s; ++i) {
                QString str = mo->getMAlignment().alignedSeqs[i].name;
                names.insert(str.replace('_', ' '));
            }

            foreach(PhyTreeObject* phyObj, phyObjs) {
                if (PhyTreeUtils::getNumSeqsFromNode(phyObj->getTree()->rootNode, names) == mo->getMAlignment().getNumSequences()) {
                    phyObj->addObjectRelation(mo, MSAEditor::MSARelation);
                    mo->addObjectRelation(phyObj, MSAEditor::TreeRelation);
                } else {
                    phyObj->removeObjectRelation(GObjectRelation(mo, MSAEditor::MSARelation));
                    mo->removeObjectRelation(GObjectRelation(phyObj, MSAEditor::TreeRelation));
                }
            }
        }
    }

    int views = 0;

    foreach(MAlignmentObject *obj, msaObjs) {
        if (views < MAX_VIEWS) {
            ++views;
            QString viewName = GObjectViewUtils::genUniqueViewName(obj->getDocument(), obj);
            log.details(tr("openining_msa_editor_for_object_%1").arg(obj->getGObjectName()));

            MSAEditor* v = new MSAEditor(viewName, obj);
            GObjectViewWindow* w = new GObjectViewWindow(v, viewName, false);
            mdiManager->addMDIWindow(w);
            msaEditors += v;
        } else
            break;
    }

    foreach(PhyTreeObject *obj, phyObjs) {
        bool found = false;
        foreach(MSAEditor *ed, msaEditors) {
            if (ed->getMSAObject() != NULL && ed->getMSAObject()->hasObjectRelation(obj, MSAEditor::TreeRelation)) {
                found = true;
                ed->addObject(obj);
                break;
            }
        }
        foreach(MSAEditor *ed, oldMsaEditors) {
            if (ed->getMSAObject() != NULL && ed->getMSAObject()->hasObjectRelation(obj, MSAEditor::TreeRelation) &&
            (ed->getPhyObject() == NULL || obj->hasObjectRelation(ed->getMSAObject(), MSAEditor::ActiveMSARelation))) {
                QString phyName = obj->getDocument()->getName();
                int i = phyName.lastIndexOf('.');
                if (i != -1) {
                    phyName = phyName.left(i);
                }
                QString msaName = ed->getMSAObject()->getDocument()->getName();
                i = msaName.lastIndexOf('.');
                if (i != -1) {
                    msaName = msaName.left(i);
                }
                i = msaName.length();
                if (phyName.left(i) == msaName) {
                    if (phyName.length() > i && phyName[i] == '_') {
                        phyName.mid(i + 1).toInt(&found);
                    }
                    if (phyName.length() == i) {
                        found = true;
                    }
                    if (found) {
                        ed->addObject(obj);
                        break;
                    }
                }
            }
        }

        if (!found && views < MAX_VIEWS) {
            ++views;
            QString viewName = GObjectViewUtils::genUniqueViewName(obj->getDocument(), obj);
            log.details(tr("openining_msa_editor_for_object_%1").arg(obj->getGObjectName()));

            MSAEditor* v = new MSAEditor(viewName, obj);
            GObjectViewWindow* w = new GObjectViewWindow(v, viewName, false);
            mdiManager->addMDIWindow(w);
        }
    }

    return ReportResult_Finished;
}

//////////////////////////////////////////////////////////////////////////
// open view from state


OpenSavedMSAEditorTask::OpenSavedMSAEditorTask(const QString& viewName, const QVariantMap& stateData) 
: ObjectViewTask(MSAEditorFactory::ID, viewName, stateData)
{
	MSAEditorState state(stateData);
	GObjectReference msaRef = state.getMSAObject();
	GObjectReference phyRef = state.getPhyObject();
	Document* msaDoc = msaRef.isValid() ? AppContext::getProject()->findDocumentByURL(msaRef.docUrl) : NULL;
	Document* phyDoc = phyRef.isValid() ? AppContext::getProject()->findDocumentByURL(phyRef.docUrl) : NULL;
	if (msaDoc == NULL && phyDoc == NULL) {
		stateIsIllegal = true;
        if (msaRef.isValid() && phyRef.isValid()) {
            stateInfo.setError(ObjectViewTask::tr("Documents %1 and %2 not found").arg(msaRef.docUrl).arg(phyRef.docUrl));
	    	return;
        }
        if (msaRef.isValid() || phyRef.isValid()) {
            stateInfo.setError(ObjectViewTask::tr("Document %1 not found").arg(msaRef.isValid() ? msaRef.docUrl : phyRef.docUrl));
	    	return;
        }
        stateInfo.setError(ObjectViewTask::tr("Invalid state"));
        return;
	}
	if (msaDoc && !msaDoc->isLoaded()) {
		documentsToLoad.append(msaDoc);
	}
	if (phyDoc && !phyDoc->isLoaded()) {
		documentsToLoad.append(phyDoc);
	}
}

void OpenSavedMSAEditorTask::open() {
    if (stateInfo.hasErrors()) {
        return;
    }
    MSAEditorState state(stateData);
    GObjectReference msaRef = state.getMSAObject();
    GObjectReference phyRef = state.getPhyObject();
    if (!msaRef.isValid() && !phyRef.isValid()) {
        stateInfo.setError(ObjectViewTask::tr("Invalid state"));
        return;
    }
    Document* msaDoc = msaRef.isValid() ? AppContext::getProject()->findDocumentByURL(msaRef.docUrl) : NULL;
    Document* phyDoc = phyRef.isValid() ? AppContext::getProject()->findDocumentByURL(phyRef.docUrl) : NULL;
	if (msaDoc == NULL && phyDoc == NULL) {
        stateIsIllegal = true;
        if (msaRef.isValid() && phyRef.isValid()) {
            stateInfo.setError(ObjectViewTask::tr("Documents %1 and %2 not found").arg(msaRef.docUrl).arg(phyRef.docUrl));
	    	return;
        }
        if (msaRef.isValid() || phyRef.isValid()) {
            stateInfo.setError(ObjectViewTask::tr("Document %1 not found").arg(msaRef.isValid() ? msaRef.docUrl : phyRef.docUrl));
	    	return;
        }
    }
    GObject* msaObj = msaRef.isValid() ? msaDoc->findGObjectByName(msaRef.objName) : NULL;
    GObject* phyObj = phyRef.isValid() ? phyDoc->findGObjectByName(phyRef.objName) : NULL;
    if (msaObj != NULL && msaObj->getGObjectType() != GObjectTypes::MULTIPLE_ALIGNMENT)
        msaObj = NULL;
    if (phyObj != NULL && phyObj->getGObjectType() != GObjectTypes::PHYLOGENETIC_TREE)
        phyObj = NULL;
    if (msaObj == NULL && phyObj == NULL) {
        stateIsIllegal = true;
        if (msaRef.isValid() && phyRef.isValid()) {
            stateInfo.setError(tr("DNA and phylogenetic objects %1 and %2 not found").arg(msaRef.objName).arg(phyRef.objName));
            return;
        }
        if (msaRef.isValid()) {
            stateInfo.setError(tr("DNA object %1 not found").arg(msaRef.objName));
            return;
        }
        if (phyRef.isValid()) {
            stateInfo.setError(tr("Phylogenetic tree object %1 not found").arg(phyRef.objName));
            return;
        }
    }
    MAlignmentObject* msaObject = NULL;
    PhyTreeObject* phyObject = NULL;
    if (msaObj)
        msaObject = qobject_cast<MAlignmentObject*>(msaObj);
    if (phyObj)
        phyObject = qobject_cast<PhyTreeObject*>(phyObj);
    assert((msaObj == NULL || msaObject != NULL) && (phyObj == NULL || phyObject != NULL));

    QSet<GObject*> allMsaObjects, allPhyObjects;
    const QList<Document*> &docs = AppContext::getProject()->getDocuments();

    foreach(Document* doc, docs) {
        allMsaObjects += doc->findGObjectByType(GObjectTypes::MULTIPLE_ALIGNMENT, UOF_LoadedOnly).toSet();
    }

    foreach(Document* doc, docs) {
        allPhyObjects += doc->findGObjectByType(GObjectTypes::PHYLOGENETIC_TREE, UOF_LoadedOnly).toSet();
    }

    if (phyObject != NULL)
        allPhyObjects.insert(phyObject);

    if (msaObject != NULL) {
        QSet<QString> names;
        for (int i = 0, s = msaObject->getMAlignment().getNumSequences(); i < s; ++i) {
            QString str = msaObject->getMAlignment().alignedSeqs[i].name;
            names.insert(str.replace('_', ' '));
        }

        foreach(GObject* obj, allPhyObjects) {
            PhyTreeObject *phyObj = qobject_cast<PhyTreeObject*>(obj);
            if (PhyTreeUtils::getNumSeqsFromNode(phyObj->getTree()->rootNode, names) == msaObject->getMAlignment().getNumSequences()) {
                phyObj->addObjectRelation(msaObject, MSAEditor::MSARelation);
                msaObject->addObjectRelation(phyObj, MSAEditor::TreeRelation);
            } else {
                phyObj->removeObjectRelation(GObjectRelation(msaObject, MSAEditor::MSARelation));
                msaObject->removeObjectRelation(GObjectRelation(phyObj, MSAEditor::TreeRelation));
            }
        }
    }

    if (phyObject != NULL) {
        foreach(GObject* obj, allMsaObjects) {
            MAlignmentObject *mo = qobject_cast<MAlignmentObject*>(obj);
            QSet<QString> names;
            for (int i = 0, s = mo->getMAlignment().getNumSequences(); i < s; ++i) {
                QString str = mo->getMAlignment().alignedSeqs[i].name;
                names.insert(str.replace('_', ' '));
            }

            if (PhyTreeUtils::getNumSeqsFromNode(phyObject->getTree()->rootNode, names) == mo->getMAlignment().getNumSequences()) {
                phyObject->addObjectRelation(mo, MSAEditor::MSARelation);
                mo->addObjectRelation(phyObject, MSAEditor::TreeRelation);
            } else {
                phyObject->removeObjectRelation(GObjectRelation(mo, MSAEditor::MSARelation));
                mo->removeObjectRelation(GObjectRelation(phyObject, MSAEditor::TreeRelation));
            }
        }
    }

    MSAEditor* v = new MSAEditor(viewName, msaObject, phyObject);
    GObjectViewWindow* w = new GObjectViewWindow(v, viewName, true);
    MWMDIManager* mdiManager = AppContext::getMainWindow()->getMDIManager();
    mdiManager->addMDIWindow(w);
	
    updateRanges(stateData, v);
}

void OpenSavedMSAEditorTask::updateRanges(const QVariantMap& stateData, MSAEditor* ctx) {
    Q_UNUSED(ctx);
	MSAEditorState state(stateData);
	int x = state.getX();
	if (x >= 0) {
		//todo:
	}

	int y = state.getY();
	if (y>=0) {
		//todo:
	}
}


//////////////////////////////////////////////////////////////////////////
// update
UpdateMSAEditorTask::UpdateMSAEditorTask(GObjectView* v, const QString& stateName, const QVariantMap& stateData) 
: ObjectViewTask(v, stateName, stateData)
{
}

void UpdateMSAEditorTask::update() {
	if (view.isNull() || view->getFactoryId() != MSAEditorFactory::ID) {
		return; //view was closed;
	}

	MSAEditor* msaView = qobject_cast<MSAEditor*>(view.data());
    assert(msaView!=NULL);
	
	OpenSavedMSAEditorTask::updateRanges(stateData, msaView);
}

} // namespace
