/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "MSAEditorTreeViewer.h"
#include "MSAEditor.h"
#include "MSAEditorSequenceArea.h"
#include "GraphicsBranchItem.h"
#include "GraphicsButtonItem.h"

#include <core_api/AppContext.h>
#include <core_api/Counter.h>
#include <core_api/DocumentModel.h>
#include <core_api/DocumentFormats.h>
#include <core_api/GUrlUtils.h>
#include <core_api/IOAdapter.h>
#include <core_api/ProjectModel.h>
#include <core_api/DocumentUtils.h>
#include <util_gui/DialogUtils.h>
#include <phyltree/CreatePhyTreeDialogController.h>
#include <phyltree/PhyTreeGeneratorRegistry.h>

#include <QtCore/QStack>

#include <QtGui/QVBoxLayout>
#include <QtGui/QMouseEvent>
#include <QtGui/QPrinter>
#include <QtGui/QPrintDialog>
#include <QtGui/QPainter>
#include <QtGui/QFileDialog>
#include <QtGui/QMessageBox>
#include <QtXml/QtXml>
#include <QtSvg/QSvgGenerator>

namespace GB2 {

#define IMAGE_DIR "image"

const int MSAEditorTreeViewer::INITIAL_WIDTH = 420;
const qreal MSAEditorTreeViewer::ZOOM_COEF = 1.2;
const qreal MSAEditorTreeViewer::MINIMUM_ZOOM = 0.4;
const qreal MSAEditorTreeViewer::MAXIMUM_ZOOM = 2.5;

const char* MSAEditorTreeViewer::IMAGE_FILTERS =
    "BMP - Windows Bitmap (*.bmp);;"
    "GIF - Graphic Interchange Format (*.gif);;"
    "JPG/JPEG format (*.jpg);;"
    "PBM - Portable Bitmap (*.pbm);;"
    "PNG - Portable Network Graphics (*.png);;"
    "PPM - Portable Pixmap (*.ppm);;"
    "TIFF - Tagged Image File format (*.tif);;"
    "XBM - X11 Bitmap (*.xbm);;"
    "XPM - X11 Pixmap (*.xpm)";

MSAEditorTreeViewer::MSAEditorTreeViewer(MSAEditorUI* _ui, QScrollBar *vb)
: editor(_ui->editor), ui(_ui), root(NULL), treeGeneratorTask(NULL), 
maxNameWidth(0), vertScrollBar(vb), zoom(1.0) 
{
    showNameLabels = editor->getMSAObject() == NULL;
    showDistanceLabels = true;
    contEnabled = false;
    bindEnabled = true;

    connect(editor, SIGNAL(si_buildStaticToolbar(GObjectView*, QToolBar*)), SLOT(sl_buildStaticToolbar(GObjectView*, QToolBar*)));
    connect(editor, SIGNAL(si_fontChanged(const QFont&)), SLOT(sl_fontChanged()));
    if (editor->getMSAObject()) {
        connect(verticalScrollBar(), SIGNAL(valueChanged(int)), SLOT(sl_scrollBarMoved(int)));
        connect(ui->seqArea, SIGNAL(si_startChanged(const QPoint& , const QPoint& )), SLOT(sl_startChanged(const QPoint& , const QPoint&)));
        connect(ui->seqArea, SIGNAL(si_selectionChanged(const MSAEditorSelection&, const MSAEditorSelection&)), SLOT(sl_selectionChanged(const MSAEditorSelection&, const MSAEditorSelection&)));
        connect(editor->getMSAObject(), SIGNAL(si_alignmentChanged(const MAlignment&, const MAlignmentModInfo&)),
            SLOT(sl_alignmentChanged(const MAlignment&, const MAlignmentModInfo&)) );
        connect(ui->getUndoAction(), SIGNAL(triggered()), SLOT(sl_recomputeView()));
        connect(ui->getRedoAction(), SIGNAL(triggered()), SLOT(sl_recomputeView()));
    }
    setFrameShape(QFrame::NoFrame);

    setAlignment(Qt::AlignRight | Qt::AlignTop);
    setTransformationAnchor(QGraphicsView::AnchorUnderMouse);
    setScene(new QGraphicsScene());

    if (editor->getPhyObject()) {
        recomputeView(true);
    }
    GCOUNTER(cvar,tvar,"MSAEditorTreeViewer");
}

void MSAEditorTreeViewer::sl_showTree() {
    if (treeGeneratorTask->getState() == Task::State_Finished) {
        if (treeGeneratorTask->hasErrors() == false) {
            if (editor->getPhyObject() == NULL) {
                QList<int> sizes = ui->splitter->sizes();
                int new_w = sizes[2] + sizes[0] - INITIAL_WIDTH;
                if (sizes[2] < 0) {
                    sizes[0] += sizes[2];
                    sizes[2] = 0;
                } else {
                    sizes[0] = INITIAL_WIDTH;
                    sizes[2] = new_w;
                }
                ui->splitter->setSizes(sizes);
            }

            if (createNewDocument) {
                const GUrl& msaURL = editor->getMSAObject()->getDocument()->getURL();
                assert(!msaURL.isEmpty());
                if (msaURL.isEmpty()) {
                    treeGeneratorTask = NULL;
                    return;
                }

                Project* p = AppContext::getProject();
                QString treeFileName = treeGeneratorTask->getSettings().fileUrl.getURLString();
                if (treeFileName.isEmpty()) {
                    treeFileName = GUrlUtils::rollFileName(msaURL.dirPath() + "//" + msaURL.baseFileName() + ".nwk", DocumentUtils::getNewDocFileNameExcludesHint());
                }
                PhyTreeObject *newObj = new PhyTreeObject(treeGeneratorTask->getResult(), "Tree");
                DocumentFormat* df = AppContext::getDocumentFormatRegistry()->getFormatById(BaseDocumentFormats::NEWICK);
                IOAdapterFactory *iof = AppContext::getIOAdapterRegistry()->getIOAdapterFactoryById(BaseIOAdapters::LOCAL_FILE);
                Document *d = new Document(df, iof, treeFileName, QList<GObject*>()<<newObj);
                p->addDocument(d);
                editor->setTree(newObj);
            } else {
                PhyTreeObject *oldObj = editor->getPhyObject();
                QString name = oldObj->getGObjectName();
                PhyTreeObject *newObj = new PhyTreeObject(treeGeneratorTask->getResult(), name);
                Document *d = oldObj->getDocument();
                d->addObject(newObj);
                editor->setTree(newObj);
                d->removeObject(oldObj);
            }
            recomputeView(true);
        }
        treeGeneratorTask = NULL;
    }
}

void MSAEditorTreeViewer::widthChanged(qreal newWidth, qreal oldWidth, bool keepLabelsWidth) {
    scale(newWidth / oldWidth, 1.0);
    if (keepLabelsWidth) {
        qreal k = oldWidth / newWidth;
        maxNameWidth *= k;
        QStack<GraphicsBranchItem*> stack;
        stack.push(root);
        while (!stack.isEmpty()) {
            GraphicsBranchItem* gbi = stack.pop();
            foreach (QGraphicsItem* item, gbi->childItems()) {
                GraphicsBranchItem* _gbi = dynamic_cast<GraphicsBranchItem*>(item);
                if (_gbi != NULL) {
                    stack.push(_gbi);
                } else {
                    item->scale(k, 1.0);
                }
            }
            if (gbi->distanceText != NULL) {
                gbi->distanceText->setPos(-(gbi->getWidth() + k * gbi->distanceText->boundingRect().width()) * 0.5, gbi->distanceText->pos().y());
            }
        }
    }
}

void MSAEditorTreeViewer::recomputeView(bool treeChanged) {
    if (root) {
        scene()->removeItem(root);
        delete root;
        root = NULL;
    }

    if (treeChanged) {
        updateAligns();
    }

    BranchCreationTask *crt = new BranchCreationTask(editor->getPhyObject()->getTree()->rootNode, editor->getFont(), aligns, editor->getRowHeight());
    connect(crt, SIGNAL(si_calculated(GraphicsBranchItem*)), SLOT(sl_viewRecomputed(GraphicsBranchItem*)));
    TaskScheduler* scheduler = AppContext::getTaskScheduler();
    scheduler->registerTopLevelTask(crt);
}

void MSAEditorTreeViewer::updateAligns() {
    int cur = 0;
    QMap<QString, int> names;
    QStack<PhyNode*> stack;
    stack.push(editor->getPhyObject()->getTree()->rootNode);

    if (editor->getMSAObject() != NULL) {
        const MAlignment &ma = editor->getMSAObject()->getMAlignment();
        int size = ma.getNumSequences();
        for (int i = 0; i < size; ++i) {
            QString str = ma.alignedSeqs[i].name;
            names.insert(str.replace('_', ' '), i);
        }
    }

    maxNameWidth = 0;
    while (!stack.empty()) {
        PhyNode *node = stack.pop();
        int size = node->branches.size();
        for (int i = size - 1; i >= 0; --i) {
            if (node->branches[i]->node2 != node)
                stack.push(node->branches[i]->node2);
        }
        if (size <= 1 && node->name != "" && node->name != "ROOT") {
            if (editor->getMSAObject() != NULL) {
                QString str = node->name.replace('_', ' ');
                assert(names.contains(str));
                aligns.insert(node, names[str]);
            } else {
                aligns.insert(node, cur++);
            }

            QFont font(editor->getFont());
            font.setPointSize(font.pointSize() - 2);
            QFontMetrics fm(font);
            int w = fm.width(node->name) + 2 * GraphicsBranchItem::TEXT_SPACE;
            maxNameWidth = qMax(maxNameWidth, w);
        }
    }
}

void MSAEditorTreeViewer::sl_viewRecomputed(GraphicsBranchItem *gbi) {
    root = gbi;

    root->setVisibleW(false);
    if (ui->seqArea != NULL && bindEnabled) {
        QScrollBar* vb = verticalScrollBar();
        vb->setMinimum(0);
        translate(0, ui->seqArea->getSequenceYRange(0, true).startPos);
        oldScrollPos = vb->sliderPosition();
    }
    scene()->addItem(root);

	if (editor->getMSAObject() != NULL) {
        if (parentWidget()->isHidden())
            showTreeAction->trigger();
    }
    rowHeight = editor->getRowHeight();

    showLabels(LabelTypes(LabelType_SequnceName | LabelType_Distance));
    updateToolBar();
    updateRect();

    if (contEnabled) {
        contEnabled = false;
        sl_contTriggered(true);
    }

//    widthChanged(ui->splitter->sizes()[0], scene()->sceneRect().width() - (showNameLabels ? maxNameWidth : 0), true);
}

void MSAEditorTreeViewer::wheelEvent(QWheelEvent *we) {
    if ((zoom > MINIMUM_ZOOM || we->delta() > 0) && (zoom < MAXIMUM_ZOOM || we->delta() < 0)) {
        if (bindEnabled && editor->getMSAObject() != NULL) {
            if (we->delta() > 0) {
                editor->increaseFont();
            } else {
                editor->decreaseFont();
            }
            oldScrollPos = vertScrollBar->sliderPosition() * editor->getRowHeight();
            verticalScrollBar()->setSliderPosition(oldScrollPos);
        } else {
            qreal zoom1 = pow(ZOOM_COEF, we->delta() / 120.0);
            zoom *= zoom1;
            scale(zoom1, zoom1);
        }
    }
    QWidget::wheelEvent(we);
}

void MSAEditorTreeViewer::mousePressEvent(QMouseEvent *e) {
    if (root) {
        if (!(e->modifiers() & Qt::ShiftModifier)) {
            if (e->button() == Qt::LeftButton) {
                root->setSelected(false);
            }
            if (dynamic_cast<GraphicsButtonItem*>(itemAt(e->pos())) == NULL) {
                setDragMode(QGraphicsView::ScrollHandDrag);
            }
        }
    }
    QGraphicsView::mousePressEvent(e);
}

void MSAEditorTreeViewer::mouseReleaseEvent(QMouseEvent *e) {
    if (root) {
        setDragMode(QGraphicsView::NoDrag);
    }
    QGraphicsView::mouseReleaseEvent(e);
}

void MSAEditorTreeViewer::sl_recomputeView() {
    if (root) {
        recomputeView(true);
    }
}

void MSAEditorTreeViewer::sl_fontChanged() {
    if (root) {
        qreal zoom1 = editor->getRowHeight() / (qreal)rowHeight;
        scale(zoom1, zoom1);
    }
    rowHeight = editor->getRowHeight();
}

void MSAEditorTreeViewer::sl_scrollBarMoved(int pos) {
    verticalScrollBar()->setMinimum(0);
    if (bindEnabled && qAbs(pos - oldScrollPos) >= rowHeight) {
        int d = (pos - oldScrollPos) / rowHeight;
        vertScrollBar->setSliderPosition(vertScrollBar->sliderPosition() + d);
    }
}

void MSAEditorTreeViewer::sl_startChanged(const QPoint& p, const QPoint& prev) {
    if (root != NULL && ui->seqArea != NULL && bindEnabled) {
        int d = ui->seqArea->getSequenceYRange(prev.y(), true).startPos - ui->seqArea->getSequenceYRange(p.y(), true).startPos;
        oldScrollPos -= d;
        verticalScrollBar()->setSliderPosition(oldScrollPos);
    }
}

void MSAEditorTreeViewer::sl_selectionChanged(const MSAEditorSelection& cur, const MSAEditorSelection& prev) {
    
}

void MSAEditorTreeViewer::sl_buildStaticToolbar(GObjectView*, QToolBar* tb) {
    if (editor->getMSAObject() != NULL) {
        QMenu *showMenu = new QMenu();
        showTreeAction = showMenu->addAction(tr("Show tree"));
        showTreeAction->setCheckable(true);
        showTreeAction->setChecked(editor->getPhyObject() != NULL);
        connect(showTreeAction, SIGNAL(triggered(bool)), SLOT(sl_showTreeTriggered(bool)));
        showSeqsAction = showMenu->addAction(tr("Show sequences area"));
        showSeqsAction->setCheckable(true);
        showSeqsAction->setChecked(true);
        connect(showSeqsAction, SIGNAL(triggered(bool)), SLOT(sl_showSeqsTriggered(bool)));

        showButton = new QToolButton();
        QAction *switchAction = new QAction(QIcon(":/core/images/page_refresh.png"), "", showButton);
        showButton->setPopupMode(QToolButton::InstantPopup);
        showButton->setDefaultAction(switchAction);
        showButton->setMenu(showMenu);

        tb->addSeparator();
        buildTreeAction = tb->addAction(QIcon(":/core/images/phylip.png"), tr("Build tree"));
        connect(buildTreeAction, SIGNAL(triggered()), SLOT(sl_buildTreeTriggered()));
        tb->addWidget(showButton);
        sortAction = tb->addAction(QIcon(":/core/images/sort_ascending.png"), tr("Sort tree"));
        connect(sortAction, SIGNAL(triggered()), SLOT(sl_sortTriggered()));
        bool locked = editor->getMSAObject() ? editor->getMSAObject()->isStateLocked() : false;
        sortAction->setEnabled(editor->getPhyObject() != NULL && (!locked));

        bindAction = NULL;
		//bindAction = tb->addAction(QIcon(":core/images/sync_lock.png"), tr("Bind tree and alignment"));
        //connect(bindAction, SIGNAL(triggered(bool)), SLOT(sl_bindTriggered(bool)));
        //bindAction->setCheckable(true);
    }

    QMenu *labelsMenu = new QMenu();
    nameLabelsAction = labelsMenu->addAction(tr("Show sequence names"));
    nameLabelsAction->setCheckable(true);
    nameLabelsAction->setChecked(showNameLabels);
    connect(nameLabelsAction, SIGNAL(triggered(bool)), SLOT(sl_showNameLabelsTriggered(bool)));
    distanceLabelsAction = labelsMenu->addAction(tr("Show distance labels"));
    distanceLabelsAction->setCheckable(true);
    distanceLabelsAction->setChecked(true);
    connect(distanceLabelsAction, SIGNAL(triggered(bool)), SLOT(sl_showDistanceLabelsTriggered(bool)));
    QToolButton *labelsButton = new QToolButton();
    labelsButton->setPopupMode(QToolButton::InstantPopup);
    labelsAction = new QAction(QIcon(":/core/images/text_ab.png"), "", labelsButton);
    labelsButton->setDefaultAction(labelsAction);
    labelsButton->setMenu(labelsMenu);
    tb->addWidget(labelsButton);

    printAction = tb->addAction(QIcon(":/core/images/printer.png"), tr("Print tree"));
    connect(printAction, SIGNAL(triggered()), SLOT(sl_printTriggered()));
    printAction->setEnabled(editor->getPhyObject() != NULL);

    QMenu *captureMenu = new QMenu();
    captureScreenAction = captureMenu->addAction(tr("Capture screen"));
    connect(captureScreenAction, SIGNAL(triggered()), SLOT(sl_captureScreenTriggered()));
    captureTreeAction = captureMenu->addAction(tr("Capture tree"));
    captureTreeAction->setEnabled(editor->getPhyObject() != NULL);
    connect(captureTreeAction, SIGNAL(triggered()), SLOT(sl_captureTreeTriggered()));
    exportAction = captureMenu->addAction(tr("Export tree in SVG"));
    connect(exportAction, SIGNAL(triggered()), SLOT(sl_exportTriggered()));
    exportAction->setEnabled(editor->getPhyObject() != NULL);
    QToolButton *captureButton = new QToolButton();
    captureButton->setPopupMode(QToolButton::InstantPopup);
    QAction *defaultCaptureAction = new QAction(QIcon(":/core/images/cam2.png"), "", captureButton);
    captureButton->setDefaultAction(defaultCaptureAction);
    captureButton->setMenu(captureMenu);
    tb->addWidget(captureButton);

    
    contAction = tb->addAction(tr("Align name labels"));
	contAction->setIcon(QIcon(":core/images/align_tree_labels.png"));
    connect(contAction, SIGNAL(triggered(bool)), SLOT(sl_contTriggered(bool)));
    contAction->setCheckable(true);

    updateToolBar();
}

void MSAEditorTreeViewer::fillMAlignment(PhyNode *node, MAlignment &msa) {
    if (editor->getMSAObject() == NULL)
        return;
    for (int i = 0, s = node->branches.size(); i < s; ++i) {
        if (node->branches[i]->node2 != node)
            fillMAlignment(node->branches[i]->node2, msa);
    }
    if (node->name != "" && node->name != "ROOT") {
        const MAlignment &old_msa = editor->getMSAObject()->getMAlignment();
        for (int i = 0, s = old_msa.getNumSequences(); i < s; ++i) {
            QString name = old_msa.alignedSeqs[i].name;
            if (name.replace('_', ' ') == node->name.replace('_', ' ')) {
                msa.alignedSeqs.append(old_msa.alignedSeqs[i]);
                aligns[node] = i;
                break;
            }
        }
    }
}

void MSAEditorTreeViewer::updateRect() {
    QRectF rect = root->mapToScene(root->childrenBoundingRect()).boundingRect();
    rect.setLeft(rect.left() - GraphicsBranchItem::DEFAULT_WIDTH);
    rect.setRight(rect.right() - (showNameLabels ? 0 : maxNameWidth) + GraphicsBranchItem::TEXT_SPACE);
    scene()->setSceneRect(rect);
}

void MSAEditorTreeViewer::sl_exportTriggered() {
    QString fileName, format;
    saveImageDialog("SVG - Scalable Vector Graphics (*.svg)", 0, fileName, format);
    if (!fileName.isEmpty()) {
        QRect rect = scene()->sceneRect().toRect();
        rect.moveTo(0, 0);
        QSvgGenerator generator;
        generator.setFileName(fileName);
        generator.setSize(rect.size());
        generator.setViewBox(rect);

        QPainter painter;
        painter.begin(&generator);
        paint(painter);
        painter.end();
    }
}

void MSAEditorTreeViewer::sl_bindTriggered(bool on) {
    if (on != bindEnabled) {
        bindEnabled = on;
        if (bindEnabled) {
            qreal zoom1 = 1.0 / zoom;
            scale(zoom1, zoom1);
            zoom = 1.0;
            oldScrollPos = vertScrollBar->sliderPosition() * editor->getRowHeight();
            verticalScrollBar()->setSliderPosition(oldScrollPos);
        }
    }
}

void MSAEditorTreeViewer::sl_contTriggered(bool on) {
    if (on != contEnabled) {
        contEnabled = on;
        QStack<GraphicsBranchItem*> stack;
        stack.push(root);
        while (!stack.empty()) {
            GraphicsBranchItem* item = stack.pop();
            if (item->nameText == NULL) {
                foreach (QGraphicsItem* citem, item->childItems()) {
                    GraphicsBranchItem* gbi = dynamic_cast<GraphicsBranchItem*>(citem);
                    if (gbi != NULL) {
                        stack.push(gbi);
                    }
                }
            } else {
                item->setWidth(on ? scene()->sceneRect().width() + scene()->sceneRect().left() - item->nameText->scenePos().x() - (showNameLabels ? item->nameText->boundingRect().width() + GraphicsBranchItem::TEXT_SPACE : - + GraphicsBranchItem::TEXT_SPACE) : 0);
            }
        }
    }
}

void MSAEditorTreeViewer::sl_buildTreeTriggered() {
    if (editor->getMSAObject() == NULL)
        return;
    PhyTreeGeneratorRegistry* registry = AppContext::getPhyTreeGeneratorRegistry();
    QStringList list = registry->getNameList();
    if(list.size()==0){
        QMessageBox::information(this, tr("Calculate phy tree"),
            tr("No algorithms for building phylogenetic tree are available.") );
        return;
    }        

    CreatePhyTreeDialogController dlg(editor->getMSAObject(), settings);

    int rc = dlg.exec();
    if (rc != QDialog::Accepted)
        return;
    
    if (treeGeneratorTask != NULL && !treeGeneratorTask->isCanceled() && !treeGeneratorTask->hasErrors()) {
        return;
    }
    if (editor->getPhyObject() != NULL) {
        createNewDocument = editor->getMSAObject()->hasObjectRelation(editor->getPhyObject(), MSAEditor::TreeRelation);
    } else {
        createNewDocument = true;
    }

    treeGeneratorTask = new PhyTreeGeneratorTask(editor->getMSAObject()->getMAlignment(), settings);
    connect(treeGeneratorTask, SIGNAL(si_stateChanged()), SLOT(sl_showTree()) );
    TaskScheduler* scheduler = AppContext::getTaskScheduler();
    scheduler->registerTopLevelTask(treeGeneratorTask);
}

void MSAEditorTreeViewer::showLabels(LabelTypes labelTypes) {
    if (root == NULL)
        return;

    QStack<GraphicsBranchItem*> stack;
    stack.push(root);
    while (!stack.isEmpty()) {
        GraphicsBranchItem *node = stack.pop();
        if (labelTypes.testFlag(LabelType_SequnceName)) {
            if (node->nameText != NULL) {
                node->nameText->setVisible(showNameLabels);
            }
        }
        if (labelTypes.testFlag(LabelType_Distance)) {
            if (node->distanceText != NULL) {
                node->distanceText->setVisible(showDistanceLabels);
            }
        }
        foreach (QGraphicsItem* item, node->childItems()) {
            GraphicsBranchItem *bitem = dynamic_cast<GraphicsBranchItem*>(item);
            if (bitem != NULL) {
                stack.push(bitem);
            }
        }
    }
}

void MSAEditorTreeViewer::sl_showNameLabelsTriggered(bool on) {
    if (on != showNameLabels) {
        QRectF rect = sceneRect();
        rect.setWidth(rect.width() + (on ? 1 : -1) * maxNameWidth);
        scene()->setSceneRect(rect);
        showNameLabels = on;
        showLabels(LabelType_SequnceName);

        if (contEnabled) {
            QStack<GraphicsBranchItem*> stack;
            stack.push(root);
            while (!stack.empty()) {
                GraphicsBranchItem* item = stack.pop();
                if (item->nameText == NULL) {
                    foreach (QGraphicsItem* citem, item->childItems()) {
                        GraphicsBranchItem* gbi = dynamic_cast<GraphicsBranchItem*>(citem);
                        if (gbi != NULL) {
                            stack.push(gbi);
                        }
                    }
                } else {
                    item->setWidth(item->getWidth() + (on ? 1 : -1) * (maxNameWidth - item->nameText->boundingRect().width() - 2 * GraphicsBranchItem::TEXT_SPACE));
                }
            }
        }
    }
}

void MSAEditorTreeViewer::sl_showDistanceLabelsTriggered(bool on) {
    if (on != showDistanceLabels) {
        showDistanceLabels = on;
        showLabels(LabelType_Distance);
    }
}

void MSAEditorTreeViewer::sl_showTreeTriggered(bool on) {
    showSeqsAction->setEnabled(on);
    if (on) {
        parentWidget()->show();
        ui->splitter->restoreState(state);
    } else {
        state = ui->splitter->saveState();
        parentWidget()->hide();
    }
}

void MSAEditorTreeViewer::sl_showSeqsTriggered(bool on) {
    showTreeAction->setEnabled(on);
    sl_showNameLabelsTriggered(!on);
    if (on) {
        ui->splitter->widget(1)->show();
        ui->splitter->widget(2)->show();
        ui->splitter->restoreState(state);
        vertScrollBar->show();
    } else {
        state = ui->splitter->saveState();
        ui->splitter->widget(1)->hide();
        ui->splitter->widget(2)->hide();
        vertScrollBar->hide();
    }
}

void MSAEditorTreeViewer::sl_printTriggered() {
    QPrinter printer;
    QPrintDialog dialog(&printer, this);
    if (dialog.exec() != QDialog::Accepted)
        return;

    QPainter painter(&printer);
    paint(painter);
}

void MSAEditorTreeViewer::sl_captureTreeTriggered() {
    QString fileName, format;
    saveImageDialog(IMAGE_FILTERS, 4, fileName, format);
    if (!fileName.isEmpty()) {
        QImage image(scene()->sceneRect().toRect().size(), QImage::Format_RGB32);
        image.fill(0xFFFFFFFF);
        QPainter painter(&image);
        paint(painter);
        image.save(fileName, format.toAscii().constData());
    }
}

void MSAEditorTreeViewer::paint(QPainter &painter) {
    painter.setBrush(QColor(0, 0, 0));
    painter.setFont(editor->getFont());
    scene()->render(&painter);
}

void MSAEditorTreeViewer::saveImageDialog(const QString& filters, int defaultIndex, QString &fileName, QString &format) {
    LastOpenDirHelper lod(IMAGE_DIR);
    Document *d = editor->getPhyObject() ? editor->getPhyObject()->getDocument() : editor->getMSAObject()->getDocument();
    QString defaultName = d->getName();
    int i = defaultName.lastIndexOf('.');
    if (i != -1) {
        defaultName = defaultName.left(i);
    }
    QString selectedFormat = filters.section(";;", defaultIndex, defaultIndex);
    QString initialPath = lod.dir + "/" + defaultName;
    fileName = QFileDialog::getSaveFileName(this, tr("Save As"), initialPath, filters, &selectedFormat);
    lod.url = fileName;
    if (fileName.isEmpty())
        return;

    format = selectedFormat.left(3).toLower();
    if (!fileName.endsWith("." + format)) {
        fileName.append("." + format);
        if (QFile::exists(fileName)) {
            QMessageBox::StandardButtons b = QMessageBox::warning(0, tr("Replace file"),
                tr("%1 already exists.\nDo you want to replace it?").arg(fileName),
                QMessageBox::Yes | QMessageBox::No);
            if (QMessageBox::Yes != b) {
                return;
            }
        }
    }
}

void MSAEditorTreeViewer::sl_captureScreenTriggered() {
    QString fileName, format;
    saveImageDialog(IMAGE_FILTERS, 4, fileName, format);
    if (!fileName.isEmpty())
        ui->saveScreenshot(fileName, format.toAscii().constData());
}

void MSAEditorTreeViewer::sl_sortTriggered() {
    if (!root || editor->getMSAObject() == NULL)
        return;

    MAlignment msa;
    msa.alphabet = editor->getMSAObject()->getMAlignment().alphabet;
    fillMAlignment(editor->getPhyObject()->getTree()->rootNode, msa);
    editor->getMSAObject()->setMAlignment(msa);
    updateAligns();
    recomputeView(false);
}

MSAEditorTreeViewer::~MSAEditorTreeViewer() {
    delete scene();
}

void MSAEditorTreeViewer::sl_alignmentChanged( const MAlignment& beforMA, const MAlignmentModInfo& mi) {
    if (!mi.sequenceListChanged) {
        return;
    }

    QSet<QString> names;
    for (int i = 0, s = editor->getMSAObject()->getMAlignment().getNumSequences(); i < s; ++i) {
        QString str = editor->getMSAObject()->getMAlignment().alignedSeqs[i].name;
        names.insert(str.replace('_', ' '));
    }

    QList<Document*> docs = AppContext::getProject()->getDocuments();
    foreach (Document* d, docs) {
        QList<GObject*> objs = d->getObjects();
        foreach (GObject* obj, objs) {
            if (obj->getGObjectType() == GObjectTypes::PHYLOGENETIC_TREE) {
                PhyTreeObject *phyObj = qobject_cast<PhyTreeObject*>(obj);
                if (PhyTreeUtils::getNumSeqsFromNode(phyObj->getTree()->rootNode, names) == editor->getMSAObject()->getMAlignment().getNumSequences()) {
                    phyObj->addObjectRelation(editor->getMSAObject(), MSAEditor::MSARelation);
                    editor->getMSAObject()->addObjectRelation(phyObj, MSAEditor::TreeRelation);
                } else {
                    phyObj->removeObjectRelation(GObjectRelation(editor->getMSAObject(), MSAEditor::MSARelation));
                    editor->getMSAObject()->removeObjectRelation(GObjectRelation(phyObj, MSAEditor::TreeRelation));
                }
            }
        }
    }

    updateToolBar();
}

void MSAEditorTreeViewer::setTree(PhyTreeObject* obj) {
    if (obj->hasObjectRelation(editor->getMSAObject(), MSAEditor::MSARelation)) {
        recomputeView(true);
    }
}

void MSAEditorTreeViewer::resizeEvent(QResizeEvent *e) {
    if (e->size().width() > 0 && e->oldSize().width() > 0) {
//        widthChanged(e->size().width(), e->oldSize().width(), true);
    }
    QGraphicsView::resizeEvent(e);
}

void MSAEditorTreeViewer::updateToolBar() {
    bool hasObj = editor->getPhyObject() != NULL;
    if (editor->getMSAObject() != NULL) {
        bool hasRel = hasObj && editor->getMSAObject()->hasObjectRelation(editor->getPhyObject(), MSAEditor::TreeRelation);
        sortAction->setEnabled(!editor->getMSAObject()->isStateLocked() && hasRel);
        showButton->setEnabled(hasObj);
        printAction->setEnabled(hasRel);
        captureTreeAction->setEnabled(hasRel);
        captureScreenAction->setEnabled(true);
		if (bindAction!=NULL) {
			bindAction->setEnabled(hasObj);
			bindAction->setChecked(hasObj && bindEnabled);
		}
    } else {
        printAction->setEnabled(hasObj);
        captureTreeAction->setEnabled(hasObj);
        captureScreenAction->setEnabled(false);
    }
    nameLabelsAction->setEnabled(hasObj);
    distanceLabelsAction->setEnabled(hasObj);
    labelsAction->setEnabled(hasObj);
    contAction->setEnabled(hasObj);
    exportAction->setEnabled(hasObj);
}

}//namespace
