/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "CircularItems.h"
#include "CircularView.h"
#include <gobjects/DNASequenceObject.h>
#include <gobjects/AnnotationTableObject.h>
#include <gobjects/AnnotationSettings.h>
#include <util_gui/GraphUtils.h>
#include <core_api/AppContext.h>
#include <QtGui/QGraphicsLineItem>

namespace GB2 {

/************************************************************************/
/* CircularAnnotationItem                                               */
/************************************************************************/

CircularAnnotationItem::CircularAnnotationItem(Annotation* ann, CircurlarAnnotationRegionItem* region) {
	assert(region->parent == NULL);
	region->parent = this;
	regions.append(region);
	this->annotation = ann;
	_boundingRect = region->boundingRect();
	isSelected = false;
}

CircularAnnotationItem::CircularAnnotationItem(Annotation* ann, QList<CircurlarAnnotationRegionItem*>& _regions) : regions(_regions) {
	this->annotation = ann;
	isSelected = false;
	QPainterPath path;
	foreach(CircurlarAnnotationRegionItem* item, regions) {
		assert(item->parent == NULL);
		item->parent = this;
		path = path.united(item->path());
	}
	_boundingRect = path.boundingRect();
}


CircularAnnotationItem::~CircularAnnotationItem()
{
	foreach(CircurlarAnnotationRegionItem* item, regions) {
		delete item->label;
		delete item;
	}
	regions.clear();
}

void CircularAnnotationItem::paint(QPainter *p,const QStyleOptionGraphicsItem *item,QWidget *widget) {
	Q_UNUSED(item);
	AnnotationSettingsRegistry* asr = AppContext::getAnnotationsSettingsRegistry();
	AnnotationSettings* as = asr->getAnnotationSettings(annotation->getAnnotationName());
	this->color = as->color;
	foreach(CircurlarAnnotationRegionItem* item, regions) {
		item->paint(p, NULL, widget);
	}
}

QRectF CircularAnnotationItem::boundingRect() const{
	return _boundingRect;
}

Annotation* CircularAnnotationItem::getAnnotation() const
{
	return annotation;
}

bool CircularAnnotationItem::contains(const QPointF &point) {
	if(_boundingRect.contains(point)) {
		foreach(CircurlarAnnotationRegionItem* item, regions)
			if(item->contains(point)) return true;
	}
	return false;
}

int CircularAnnotationItem::containsRegion(const QPointF &point) {
	if(_boundingRect.contains(point)) {
		foreach(CircurlarAnnotationRegionItem* item, regions) {
			if(item->contains(point)) 
				return item->number;
		}
	}
	return -1;
}

void CircularAnnotationItem::setSelected( bool isSelected )
{
	this->isSelected = isSelected;
}

const QList<CircurlarAnnotationRegionItem*>& CircularAnnotationItem::getRegions()
{
	return regions;
}
/************************************************************************/
/* CircurlarAnnotationRegionItem                                        */
/************************************************************************/

CircurlarAnnotationRegionItem::CircurlarAnnotationRegionItem(const QPainterPath& path, bool _isShort, int _number ) 
	:QGraphicsPathItem(path), parent(NULL), label(NULL), number(_number), isShort(_isShort)
{
}

void CircurlarAnnotationRegionItem::paint( QPainter *p, const QStyleOptionGraphicsItem *option, QWidget *widget /* = 0 */ )
{
	Q_UNUSED(option);
	assert(parent!=NULL);
	QPen pen(Qt::black);
	pen.setWidth(1);
	if(parent->isSelected) {
		pen.setWidth(2);
	}
	
	p->setPen(pen);
	CircularViewRenderArea* renderArea = static_cast<CircularViewRenderArea*>(widget);
	int yLevel = renderArea->annotationYLevel[parent->annotation];
	int innerRadius = CircularViewRenderArea::INNER_ELLIPSE_SIZE/2 + yLevel * CircularViewRenderArea::ELLIPSE_DELTA/2;
	int outerRadius = CircularViewRenderArea::OUTER_ELLIPSE_SIZE/2 + yLevel * CircularViewRenderArea::ELLIPSE_DELTA/2;
	QColor &color(parent->color);
	QRadialGradient radialGrad(QPointF(0, 0), outerRadius );
	radialGrad.setColorAt(1, color);
	radialGrad.setColorAt((float)innerRadius / (float)outerRadius, QColor(color.red()*0.7,color.green()*0.7,color.blue()*0.7));
	radialGrad.setColorAt(0, Qt::black);
	
	p->fillPath(path(), radialGrad);

	if(!isShort || parent->isSelected)
		p->drawPath(path());
}

CircularAnnotationLabel* CircurlarAnnotationRegionItem::getLabel() const
{
	return label;
}

void CircurlarAnnotationRegionItem::setLabel( CircularAnnotationLabel* label )
{
	this->label = label;
}
/************************************************************************/
/* CircularAnnotationLabel                                              */
/************************************************************************/

static bool labelAngleMoreThan(CircularAnnotationLabel* l1, CircularAnnotationLabel* l2) {
	return l1->getLabelY() > l2->getLabelY();
}

static bool labelLengthLessThan(CircularAnnotationLabel* l1, CircularAnnotationLabel* l2) {
	int length1 =  l1->getAnnotation()->getLocation()[l1->getRegion()].len;
	int length2 =  l2->getAnnotation()->getLocation()[l2->getRegion()].len;
	return length1 < length2;
}

CircularAnnotationLabel::CircularAnnotationLabel( Annotation* ann, int _region, int sequenceLength, const QFont& font, CircularViewRenderArea* renderArea ) : 
annotation(ann), labelFont(font), region(_region)
{
	const LRegion& r = annotation->getLocation()[region];
	float startAngle = (float)r.startPos / (float)sequenceLength * 360;
	float spanAngle = (float)r.len / (float)sequenceLength * 360;
	float middleAngle = 360 - (startAngle + spanAngle /2.0);
	annotationAngle = middleAngle;
	const int yLevel = renderArea->annotationYLevel[annotation];
	const int count = renderArea->regionY.count();
	outerRect = QRectF(-CircularViewRenderArea::OUTER_ELLIPSE_SIZE/2 - count * CircularViewRenderArea::ELLIPSE_DELTA, 
		-CircularViewRenderArea::OUTER_ELLIPSE_SIZE - count * CircularViewRenderArea::ELLIPSE_DELTA, 
		CircularViewRenderArea::OUTER_ELLIPSE_SIZE + count * CircularViewRenderArea::ELLIPSE_DELTA*2,
		CircularViewRenderArea::OUTER_ELLIPSE_SIZE + count * CircularViewRenderArea::ELLIPSE_DELTA*2);	

	innerRect = QRectF(-CircularViewRenderArea::MIDDLE_ELLIPSE_SIZE/2 - yLevel * CircularViewRenderArea::ELLIPSE_DELTA/2,
		-CircularViewRenderArea::MIDDLE_ELLIPSE_SIZE/2 - yLevel * CircularViewRenderArea::ELLIPSE_DELTA/2, 
		CircularViewRenderArea::MIDDLE_ELLIPSE_SIZE + yLevel * CircularViewRenderArea::ELLIPSE_DELTA,
		CircularViewRenderArea::MIDDLE_ELLIPSE_SIZE + yLevel * CircularViewRenderArea::ELLIPSE_DELTA);
	labelY = sin(middleAngle / 180.0 * Pi) * innerRect.height() / 2;
	setVisible(false);
}

void CircularAnnotationLabel::prepareLabels(QList<CircularAnnotationLabel*>& labelItems, const CircularViewRenderArea* renderArea) {
	
	QList<CircularAnnotationLabel*> left, right;

	if(labelItems.count() > 200) {
		QList<CircularAnnotationLabel*> newList;
		qSort<QList<CircularAnnotationLabel*>::iterator>(labelItems.begin(), labelItems.end(), labelLengthLessThan);
		int l_j = 0;
		int r_j = 0;

		ADVSequenceObjectContext* ctx = renderArea->view->getSequenceContext();
		const int seqLen = ctx->getSequenceLen();
		float dAlpha = 360 * CircularViewRenderArea::ARROW_LENGTH / (float)Pi / (CircularViewRenderArea::OUTER_ELLIPSE_SIZE + CircularViewRenderArea::INNER_ELLIPSE_SIZE);

		for(QList<CircularAnnotationLabel*>::iterator i = labelItems.begin(); i != labelItems.end() && (r_j < 100 || l_j < 100); ++i) {
			CircularAnnotationLabel* label = *i;
			const LRegion &r = label->annotation->getLocation()[label->region];
			if(r.len / (float)seqLen * 360 >= dAlpha) {
				newList << label;
				if(label->annotationAngle < 90 || label->annotationAngle > 270) {
					right << label;
					r_j++;
				}
				else {
					left << label;
					l_j++;
				}
				
			}
		}
		labelItems.clear();
		labelItems.append(newList);
	} else {
		for(int i=0;i<labelItems.count();i++) {
			if(labelItems[i]->annotationAngle < 90 || labelItems[i]->annotationAngle > 270)
				right << labelItems[i];
			else 
				left << labelItems[i];
		}
	}

	//qSort<QList<CircularAnnotationLabel*>::iterator>(labelItems.begin(), labelItems.end(), labelAngleMoreThan);
	qSort<QList<CircularAnnotationLabel*>::iterator>(right.begin(), right.end(), labelAngleMoreThan);
	qSort<QList<CircularAnnotationLabel*>::iterator>(left.begin(), left.end(), labelAngleMoreThan);
	
	foreach(CircularAnnotationLabel* l, labelItems) l->setVisible(true);
	
	const int minDelta = 15;
	
	for(int k = 0; k < 2; k++) {
		QList<CircularAnnotationLabel*>& side = k==0 ? left : right;
		for(int i=1;i<side.count();i++) {
			float& prev = side[i-1]->labelY;
			float& current = side[i]->labelY;
			float delta =  prev - current;
			float shift = 0;
			if(delta > 0) {
				if(delta < minDelta) {
					shift = minDelta - delta/2;
				}
			} else {
				shift = minDelta - delta;
			}
			current -= shift;
		}
		if(!side.isEmpty()) {
			const CircularAnnotationLabel* last = side.last();
			if(last->labelY < 0) {
				float globalShift = -last->labelY - last->innerRect.height()/2;
				for(int i=side.count() - 1;i > 0 && globalShift > 0;i--) {
					float& prev = side[i]->labelY;
					float& current = side[i-1]->labelY;
					float delta =  (current - prev) - minDelta;
					if(delta > 0) {
						if(globalShift - delta < 0)
							delta -= delta - globalShift;
						for(int j = side.count() - 1; j >= i; j--) 
							side[j]->labelY += delta;
						globalShift -= delta;
					}
				}
				if(globalShift > 0) {
					globalShift /= 2;
					foreach(CircularAnnotationLabel* label, side) 
						label->labelY += globalShift;
				}
			}
		}
	}
}

void CircularAnnotationLabel::paint( QPainter *p,const QStyleOptionGraphicsItem *item,QWidget *widget )
{
	Q_UNUSED(item);
	
	if(!isVisible()) return;

	CircularViewRenderArea* renderArea = static_cast<CircularViewRenderArea*>(widget);

	point1 = QPointF((outerRect.width()/2) * (annotationAngle > 90 && annotationAngle <270 ? -1 : 1),-labelY);
	point2 = QPointF (innerRect.width()/2 * cos(annotationAngle / 180.0 * Pi),-innerRect.height()/2 * sin(annotationAngle / 180.0 * Pi));
	
	QPen pen(Qt::black);
	pen.setWidth(1);
	labelFont.setBold(false);
	if(renderArea->circItems[annotation]->isSelected) {
		labelFont.setBold(true);
		p->setFont(labelFont);
		pen.setWidth(2);
	}
	p->setPen(pen);

	QRectF bound = boundingRect();

	p->fillRect(bound, Qt::white);
	p->drawText(bound, annotation->getAnnotationName());
	p->drawLine(point1, point2);
	
}

Annotation* CircularAnnotationLabel::getAnnotation() const
{
	return annotation;
}

QRectF CircularAnnotationLabel::boundingRect() const
{
	QFontMetrics fm(labelFont);
	QPointF point1((outerRect.width()/2) * (annotationAngle > 90 && annotationAngle <270 ? -1 : 1),-labelY);
	QRectF bound = fm.boundingRect(annotation->getAnnotationName());
	if(annotationAngle < 90 || annotationAngle > 270) {
		bound.moveBottomLeft(point1+ QPointF(0, bound.height()/2));
	} else {
		bound.moveBottomRight(point1 + QPointF(0, bound.height()/2));
	}
	return bound;
}

bool CircularAnnotationLabel::contains( const QPointF &point )
{
	return boundingRect().contains(point) || QGraphicsLineItem(QLine(point1.toPoint(), point2.toPoint())).contains(point);
}

bool CircularAnnotationLabel::operator<( const CircularAnnotationLabel& a )
{
	return (labelY > a.labelY);
}

float CircularAnnotationLabel::getLabelY() const
{
	return labelY;
}

int CircularAnnotationLabel::getRegion() const
{
	return region;
}
} // namespace GB2