/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "CircularView.h"

#include <core_api/DNAAlphabet.h>
#include <core_api/SelectionModel.h>
#include <core_api/Log.h>
#include <core_api/Timer.h>
#include <core_api/AppContext.h>

#include <gobjects/DNASequenceObject.h>
#include <gobjects/AnnotationTableObject.h>
#include <gobjects/AnnotationSettings.h>

#include <selection/DNASequenceSelection.h>
#include <util_gui/GraphUtils.h>
#include <util_gui/GScrollBar.h>
#include <util_text/FormatUtils.h>


#include <QtGui/QTextEdit>
#include <QtGui/QGridLayout>
#include <QtGui/QPainter>
#include <QtGui/QFontMetrics>
#include <QtGui/QDialog>

#include "CircularItems.h"

#define Pi 3.1415926535897932384626433832795	

namespace GB2 {

static LogCategory log(ULOG_CAT_ADV);

CircularView::CircularView(QWidget* p, ADVSequenceObjectContext* ctx) : GSequenceLineViewAnnotated(p, ctx)
{
	foreach(AnnotationTableObject* obj, ctx->getAnnotationObjects()) {
		registerAnnotations(obj->getAnnotations());
	}

	renderArea = new CircularViewRenderArea(this);

	renderArea->setMinimumHeight(600);
	renderArea->setMinimumWidth(600);

	connect(ctx->getSequenceGObject(), SIGNAL(si_sequenceChanged()), this, SLOT(sl_sequenceChanged()));
	pack();
}

CircularView::~CircularView() {}

QSize CircularView::sizeHint() const {
	return QSize(600, 600);
}

void CircularView::pack() {
	assert(layout() == NULL);
	setMinimumHeight(200);
	QVBoxLayout *layout = new QVBoxLayout();
	scrollArea = new QScrollArea();
	scrollArea->setWidgetResizable(true);
	scrollArea->setWidget(renderArea);
	scrollArea->setAlignment(Qt::AlignCenter);
	layout->setContentsMargins(0,0,0,0);
	layout->addWidget(scrollArea);
	setLayout(layout);
	setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
}

void CircularView::mousePressEvent(QMouseEvent * e) {
	GSequenceLineViewAnnotated::mousePressEvent(e);
	
	QPoint p = toRenderAreaPoint(e->pos());
	QPoint point(p.x() - renderArea->width()/2 , p.y() - renderArea->height()/2);
	float norm = sqrt((double)point.rx()*point.rx() + point.ry()*point.ry());
	float arcsin = asin(abs(point.ry())/norm);
	if(point.rx() < 0) arcsin = Pi - arcsin;
	if(point.ry() < 0) arcsin = 2*Pi - arcsin;
	lastPressPos = 180 * 16 * arcsin / Pi;
	lastMovePos = lastPressPos;
	lastMouseY = point.y();
	currectSelectionLen = 0;
	QWidget::mousePressEvent(e);
}

void CircularView::mouseMoveEvent( QMouseEvent * e )
{
	if (lastPressPos == -1) {
		QWidget::mouseMoveEvent(e);
		return;
	}
	if (e->buttons() & Qt::LeftButton) {
		QPoint areaPoint = toRenderAreaPoint(e->pos());
		QPoint point(areaPoint.x() - renderArea->width()/2 , areaPoint.y() - renderArea->height()/2);
		float norm = sqrt((double)point.rx()*point.rx() + point.ry()*point.ry());
		
		float arcsin = asin(abs(point.ry())/norm);
		if(point.rx() < 0) arcsin = Pi - arcsin;
		if(point.ry() < 0) arcsin = 2*Pi - arcsin;
		float a = 180 * 16 * arcsin / Pi;
		currectSelectionLen += a - lastMovePos + 1;

		// compute selection
		const int seqLen = ctx->getSequenceLen();
		int selStart = lastPressPos / 5760.0 * seqLen;
		int selLen = currectSelectionLen / 5760.0 * seqLen;

		if (selLen < 0) {
		//	if(lastMouseY < point.ry()) {
		//		selLen+=5760;
		//	} else {
				selStart = selStart + selLen;
				selLen *= -1;
		//	}
		}
		if (selLen > seqLen) {
			selLen = seqLen;
		}
		if (selStart + selLen > seqLen) {
            /*getSequenceContext()->getSequenceSelection()->clear();
              getSequenceContext()->getSequenceSelection()->addRegion(LRegion(selStart, seqLen - selStart));
              getSequenceContext()->getSequenceSelection()->addRegion(LRegion(selStart, selStart + selLen - seqLen));*/
            setSelection(LRegion(selStart, seqLen - selStart));
		} else {
			setSelection(LRegion(selStart, selLen));
		}
		lastMovePos = a;
		lastMouseY = point.y();
	}
	QWidget::mouseMoveEvent(e);
}

void CircularView::mouseReleaseEvent(QMouseEvent* e) {
	GSequenceLineViewAnnotated::mouseReleaseEvent(e);
}

void CircularView::sl_onAnnotationSelectionChanged(AnnotationSelection* selection, const QList<Annotation*>& added, const QList<Annotation*>& removed) {
	GSequenceLineViewAnnotated::sl_onAnnotationSelectionChanged(selection, added, removed);
	renderArea->update();
}

void CircularView::sl_onDNASelectionChanged( LRegionsSelection* thiz, const QList<LRegion>& added, const QList<LRegion>& removed )
{
	GSequenceLineViewAnnotated::sl_onDNASelectionChanged(thiz, added, removed);
	renderArea->update();
}

QList<AnnotationSelectionData> CircularView::selectAnnotationByCoord( const QPoint& coord ) const
{
	QList<AnnotationSelectionData> res;
	CircularViewRenderArea* renderArea = static_cast<CircularViewRenderArea*>(this->renderArea);
	QPoint cp(coord - QPoint(renderArea->width()/2, renderArea->height()/2));
	foreach(CircularAnnotationItem* item, renderArea->circItems) {
		int region = item->containsRegion(cp);
		if(region != -1) {
			res.append(AnnotationSelectionData(item->getAnnotation(), region));
			return res;
		}
	}
	foreach(CircularAnnotationItem* item, renderArea->circItems) {
		foreach(CircurlarAnnotationRegionItem* r, item->getRegions()) {
			if(r->getLabel()->isVisible() && r->getLabel()->contains(cp)) {
				res.append(AnnotationSelectionData(item->getAnnotation(), item->getRegions().indexOf(r)));
				return res;
			}
		}
	}
	return res;
}

void CircularView::resizeEvent( QResizeEvent* e )
{
	//GSequenceLineViewAnnotated::resizeEvent(e);
	updateScrollBar();
	QWidget::resizeEvent(e);
}

/************************************************************************/
/* CircularViewRenderArea                                               */
/************************************************************************/
CircularViewRenderArea::CircularViewRenderArea(CircularView* d) : GSequenceLineViewAnnotatedRenderArea(d, true), circularView(d), needResizeView(true), scale(1) {
	
}

void CircularViewRenderArea::drawAll(QPaintDevice* pd) {
	QPainter p(pd);
	p.setRenderHint(QPainter::Antialiasing);
	GSLV_UpdateFlags uf = view->getUpdateFlags();
	bool completeRedraw = uf.testFlag(GSLV_UF_NeedCompleteRedraw) || uf.testFlag(GSLV_UF_ViewResized) || 
		uf.testFlag(GSLV_UF_AnnotationsChanged);

	if (completeRedraw) {
		//scale = qMin(1.0f, qMax(0.3f,(float)parentWidget()->height() / (float)OUTER_ELLIPSE_SIZE));
		QPainter pCached(cachedView);
		pCached.setRenderHint(QPainter::Antialiasing);
		pCached.fillRect(0, 0, pd->width(), pd->height(), Qt::white);
		pCached.translate(pd->width()/2, pd->height()/2);
		//pCached.scale(scale,scale);
		pCached.setPen(Qt::black);
		drawSequence(pCached);
		drawAnnotations(pCached);
		drawRuler(pCached);
		pCached.end();
	}
	p.drawPixmap(0, 0, *cachedView);
	p.translate(width()/2, height()/2);
	//p.scale(scale,scale);


	drawAnnotationsSelection(p);

	drawSequenceSelection(p);
}

void CircularViewRenderArea::drawAnnotationsSelection(QPainter& p) {

	ADVSequenceObjectContext* ctx = view->getSequenceContext();

	if(ctx->getAnnotationsSelection()->getSelection().isEmpty()) {
		return;
	}

	foreach(const AnnotationSelectionData& asd, ctx->getAnnotationsSelection()->getSelection()) {
		AnnotationTableObject* o = asd.annotation->getGObject();
		if (ctx->getAnnotationObjects().contains(o)) {
			if(circItems.contains(asd.annotation)) {
				CircularAnnotationItem* item = circItems[asd.annotation];
				item->setSelected(true);
				item->paint(&p, NULL, this);
				foreach(const CircurlarAnnotationRegionItem* r, item->getRegions()) {
					if(r->getLabel()->isVisible()) {
						r->getLabel()->paint(&p, NULL, this);
					}
				}
				item->setSelected(false);
			}
		}
	}
}

void CircularViewRenderArea::drawSequence(QPainter& p) {
	QPen boldPen(Qt::black);
	boldPen.setWidth(3);
	ADVSequenceObjectContext* ctx = view->getSequenceContext();

	assert( ctx->getSequenceGObject() != NULL ); 
		
	QString docName = ctx->getSequenceGObject()->getDocument()->getName();
	QRectF r = p.boundingRect(0,0,1000,1000, Qt::AlignLeft, docName);
	r.moveCenter(QPointF(0,0));
	p.setPen(boldPen);
	p.drawEllipse(-SEQUENCE_ELLIPSE_SIZE/2, -SEQUENCE_ELLIPSE_SIZE/2, SEQUENCE_ELLIPSE_SIZE, SEQUENCE_ELLIPSE_SIZE);
	p.drawText(r, docName);
}

void CircularViewRenderArea::drawSequenceSelection( QPainter& p )
{
	ADVSequenceObjectContext* ctx = view->getSequenceContext();
	const int seqLen = ctx->getSequenceLen();
	const QList<LRegion>& selection = view->getSequenceContext()->getSequenceSelection()->getSelectedRegions();
	QPen selectionPen(Qt::black);
	selectionPen.setWidth(1);
	p.setPen(selectionPen);
	QColor color(Qt::blue);
	color.setAlpha(128);
	QRadialGradient radialGrad(QPointF(0, 0), SEQUENCE_ELLIPSE_SIZE/2 + ELLIPSE_DELTA);
	radialGrad.setColorAt(1, color);
	radialGrad.setColorAt((float)(SEQUENCE_ELLIPSE_SIZE/2 - ELLIPSE_DELTA) / (float)(SEQUENCE_ELLIPSE_SIZE/2 + ELLIPSE_DELTA), QColor(color.red()*0.7,color.green()*0.7,color.blue()*0.7));
	radialGrad.setColorAt(0, Qt::black);

	foreach(const LRegion& r, selection) {
		QPainterPath path;
		QRectF outerRect(-SEQUENCE_ELLIPSE_SIZE/2 - ELLIPSE_DELTA/2, -SEQUENCE_ELLIPSE_SIZE/2 - ELLIPSE_DELTA/2, SEQUENCE_ELLIPSE_SIZE + ELLIPSE_DELTA, SEQUENCE_ELLIPSE_SIZE + ELLIPSE_DELTA);
		QRectF innerRect(-SEQUENCE_ELLIPSE_SIZE/2 + ELLIPSE_DELTA/2, -SEQUENCE_ELLIPSE_SIZE/2 + ELLIPSE_DELTA/2, SEQUENCE_ELLIPSE_SIZE - ELLIPSE_DELTA, SEQUENCE_ELLIPSE_SIZE - ELLIPSE_DELTA);
		float startAngle = r.startPos / (float)seqLen * 360;
		float spanAngle = r.len / (float)seqLen * 360;
		path.moveTo(outerRect.width()/2 * cos(-startAngle / 180.0 * Pi),-outerRect.width()/2 * sin(-startAngle / 180.0 * Pi));
		path.arcTo(outerRect, -startAngle, -spanAngle);
		path.arcTo(innerRect, -startAngle-spanAngle, spanAngle);
		path.closeSubpath();
		p.drawPath(path);
		p.fillPath(path, radialGrad);
	}
}


void CircularViewRenderArea::drawRuler( QPainter& p )
{
	int labelCount = 10;
	ADVSequenceObjectContext* ctx = view->getSequenceContext();
	int len = ctx->getSequenceLen();
	
	p.setFont(rulerFont);

	for(int i=0; i< labelCount; i++) {
		float d = 2*Pi - i * 2 * Pi / labelCount;
		QPoint point1 ( RULER_ELLIPSE_SIZE * cos(d) / 2, - RULER_ELLIPSE_SIZE * sin(d) / 2 );
		QPoint point2 ( SEQUENCE_ELLIPSE_SIZE * cos(d) / 2, - SEQUENCE_ELLIPSE_SIZE * sin(d) / 2 );
		QString label = FormatUtils::formatNumber(i * len / (float)labelCount + 1);
		QRect bounding = p.boundingRect(0,0,1000,1000, Qt::AlignLeft, label); 
		d*=180/(float)Pi;
		if(d < 90) {
			bounding.moveTopRight(point1);
		} else if(d < 180) {
			bounding.moveTopLeft(point1);
		} else if(d < 270) {
			bounding.moveBottomLeft(point1);
		} else {
			bounding.moveBottomRight(point1);
		}
		
		p.drawLine(point1,point2);
		p.fillRect(bounding, Qt::white);
		p.drawText(bounding, label);
	}
}

void CircularViewRenderArea::drawAnnotations(QPainter& p) {
	ADVSequenceObjectContext* ctx = view->getSequenceContext();

	QPen pen1(Qt::SolidLine);
	pen1.setWidth(1);
	foreach(CircularAnnotationItem* item, circItems) {
		delete item;
	}
	circItems.clear();
	labelList.clear();
	annotationYLevel.clear();
	regionY.clear();
	circItems.clear();

	AnnotationSettingsRegistry* asr = AppContext::getAnnotationsSettingsRegistry();
	//for(QSet<AnnotationTableObject*>::const_iterator i = ctx->getAnnotationObjects().begin(); i != ctx->getAnnotationGObjects().constEnd(); i++) {
	//TODO: there need const order of annotation tables
	foreach(AnnotationTableObject* ao, ctx->getAnnotationObjects()) {
		foreach(Annotation* a, ao->getAnnotations()) {
			AnnotationSettings* as = asr->getAnnotationSettings(a->getAnnotationName());
			buildAnnotationItem(DrawAnnotationPass_DrawFill, a, false, as);
		}
	}
	pen1.setWidth(2);
	foreach(AnnotationTableObject* ao, ctx->getAnnotationObjects()) {
		foreach(Annotation* a, ao->getAnnotations()) {
			AnnotationSettings* as = asr->getAnnotationSettings(a->getAnnotationName());
			buildAnnotationLabel(p.font(), a, as);
		}
	}
	
	CircularAnnotationLabel::prepareLabels(labelList, this);

	//Adjust renderArea size
	QRect clipRect(-this->width()/2, -this->height()/2, this->width(), this->height());
	foreach(CircularAnnotationLabel* label, labelList) {
		QRect bound(label->boundingRect().toRect());
		if(!clipRect.contains(bound)) {
			if(clipRect.top() > bound.top()) {
				clipRect.adjust(0, -(clipRect.top() - bound.top()), 0 ,0);
			}
			if(clipRect.bottom() < bound.bottom()) {
				clipRect.adjust(0, 0, 0 , bound.bottom() - clipRect.bottom());
			}
			if(clipRect.left() > bound.left()) {
				clipRect.adjust(bound.left() - clipRect.left(), 0, 0 , 0);
			}
			if(clipRect.right() < bound.right()) {
				clipRect.adjust(0, 0, bound.right() - clipRect.right() , 0);
			}
		}
	}
	//printf("\nBefore: width: %d, height: %d\n",this->width(), this->height());
	clipRect.setWidth(qMax(clipRect.width(),600));
	clipRect.setHeight(qMax(clipRect.height(),600));
	QSize oldSize = this->size();
	if(oldSize.rwidth() != clipRect.width() || oldSize.rheight() != clipRect.height()) {
		this->setMinimumSize(clipRect.size());
		needResizeView = true;
	}
	//printf("\nClipRect: left: %d, top: %d, width: %d, height: %d\n",clipRect.left(), clipRect.top(), clipRect.width(), clipRect.height());
	//printf("\nAfter: width: %d, height: %d\n",this->width(), this->height());
	foreach(CircularAnnotationItem* item, circItems) {
		item->paint(&p, NULL, this);
	}
	foreach(CircularAnnotationLabel* label, labelList) {
		label->paint(&p, NULL, this);
	}
	
}

void CircularViewRenderArea::buildAnnotationItem(DrawAnnotationPass pass, Annotation* a, bool selected /* = false */, const AnnotationSettings* as /* = NULL */) {
	if (!as->visible && (pass == DrawAnnotationPass_DrawFill || !selected)) {
		return;
	}

	ADVSequenceObjectContext* ctx = view->getSequenceContext();

	const int seqLen = ctx->getSequenceLen();

	const QList<LRegion>& location = a->getLocation();
	
	const LRegion generalLocation(location.first().startPos, location.last().startPos - location.first().startPos + location.last().len);

	int yLevel = 0;
	bool yFind = false;
	for(;yLevel<regionY.count();yLevel++) {
		bool intersects = false;
		foreach(const LRegion& r, regionY[yLevel]) {
			if(r.intersects(generalLocation)) {
				intersects = true;
				break;
			}
		}
		if(!intersects) {
			regionY[yLevel].append(generalLocation);
			yFind = true;
			break;
		}
	}
	if(!yFind) { 
		QList<LRegion> newLevel;
		newLevel.append(generalLocation);
		regionY.append(newLevel);
	}
	annotationYLevel[a] = yLevel;

	QList<CircurlarAnnotationRegionItem*> regions;

	foreach(const LRegion& r, location) {
		float startAngle = (float)r.startPos / (float)seqLen * 360;
		float spanAngle = qMin((float)r.len / (float)seqLen * 360, 360.0f);
		QPainterPath path;
		
		QRect outerRect(-OUTER_ELLIPSE_SIZE/2 - yLevel * ELLIPSE_DELTA/2, -OUTER_ELLIPSE_SIZE/2 - yLevel * ELLIPSE_DELTA/2, OUTER_ELLIPSE_SIZE + yLevel * ELLIPSE_DELTA, OUTER_ELLIPSE_SIZE + yLevel * ELLIPSE_DELTA);
		QRect innerRect(-INNER_ELLIPSE_SIZE/2 - yLevel * ELLIPSE_DELTA/2, -INNER_ELLIPSE_SIZE/2 - yLevel * ELLIPSE_DELTA/2, INNER_ELLIPSE_SIZE + yLevel * ELLIPSE_DELTA, INNER_ELLIPSE_SIZE + yLevel * ELLIPSE_DELTA);
		QRect middleRect(-MIDDLE_ELLIPSE_SIZE/2 - yLevel * ELLIPSE_DELTA/2, -MIDDLE_ELLIPSE_SIZE/2 - yLevel * ELLIPSE_DELTA/2, MIDDLE_ELLIPSE_SIZE + yLevel * ELLIPSE_DELTA, MIDDLE_ELLIPSE_SIZE + yLevel * ELLIPSE_DELTA);
		float dAlpha = 360 * ARROW_LENGTH / (float)Pi / (OUTER_ELLIPSE_SIZE + INNER_ELLIPSE_SIZE);
		bool isShort = r.len / (float)seqLen * 360 < dAlpha;
		if(isShort) {
			path.moveTo(outerRect.width()/2 * cos(-startAngle / 180.0 * Pi),-outerRect.height()/2 * sin(-startAngle / 180.0 * Pi));
			path.arcTo(outerRect, -startAngle, -spanAngle);
			path.arcTo(innerRect, -startAngle-spanAngle, spanAngle);
			path.closeSubpath();
		} else {
			if(a->isOnComplementStrand()) {
				path.moveTo(outerRect.width()/2 * cos(-(startAngle + dAlpha) / 180.0 * Pi),-outerRect.height()/2 * sin(-(startAngle + dAlpha) / 180.0 * Pi));
				path.lineTo((outerRect.width()/2 + dAlpha) * cos(-(startAngle + dAlpha) / 180.0 * Pi),-(outerRect.width()/2 + dAlpha) * sin(-(startAngle + dAlpha) / 180.0 * Pi));
				path.lineTo(middleRect.width()/2 * cos(-startAngle / 180.0 * Pi),-middleRect.height()/2 * sin(-startAngle/ 180.0 * Pi));
				path.lineTo((innerRect.width()/2 - dAlpha) * cos(-(startAngle + dAlpha) / 180.0 * Pi),-(innerRect.width()/2 - dAlpha) * sin(-(startAngle + dAlpha) / 180.0 * Pi));
				path.arcTo(innerRect, -(startAngle + dAlpha), -(spanAngle - dAlpha));
				path.arcTo(outerRect, -startAngle - spanAngle, spanAngle - dAlpha);
				path.closeSubpath();
			} else {
				path.moveTo(outerRect.width()/2 * cos(-startAngle / 180.0 * Pi),-outerRect.height()/2 * sin(-startAngle / 180.0 * Pi));
				path.arcTo(outerRect, -startAngle, -(spanAngle - dAlpha));
				path.lineTo((outerRect.width()/2 + ARROW_HEIGHT_DELTA) * cos((-startAngle - (spanAngle - dAlpha)) / 180.0 * Pi),-(outerRect.height()/2 + ARROW_HEIGHT_DELTA) * sin((-startAngle - (spanAngle - dAlpha)) / 180.0 * Pi));
				path.lineTo(middleRect.width()/2 * cos((-startAngle - spanAngle) / 180.0 * Pi),-middleRect.height()/2 * sin((-startAngle - spanAngle)/ 180.0 * Pi));
				path.lineTo((innerRect.width()/2 - ARROW_HEIGHT_DELTA) * cos((-startAngle - (spanAngle - dAlpha)) / 180.0 * Pi),-(innerRect.height()/2 - ARROW_HEIGHT_DELTA) * sin((-startAngle - (spanAngle - dAlpha)) / 180.0 * Pi));
				path.arcTo(innerRect, -startAngle - (spanAngle - dAlpha), spanAngle - dAlpha);
				path.closeSubpath();
			}
		}
		regions.append(new CircurlarAnnotationRegionItem(path, isShort, location.indexOf(r)));
	}

	CircularAnnotationItem* item = new CircularAnnotationItem(a, regions);
	circItems[a] = item;
}

void CircularViewRenderArea::buildAnnotationLabel(const QFont& font, Annotation* a, const AnnotationSettings* as) {
	
	if (!as->visible) {
		return;
	}

	ADVSequenceObjectContext* ctx = view->getSequenceContext();

	const int seqLen = ctx->getSequenceLen();
	const QList<LRegion>& location = a->getLocation();
	for(int r=0;r<location.count();r++) {
		CircularAnnotationLabel* label = new CircularAnnotationLabel(a, r, seqLen, font, this);
		labelList.append(label);
		circItems[a]->getRegions()[r]->setLabel(label);
	}
}

LRegion CircularViewRenderArea::getAnnotationYRange(Annotation* a, const LRegion& r, const AnnotationSettings* as) const{
	Q_UNUSED(a);
	Q_UNUSED(r);
	Q_UNUSED(as);
	return LRegion(0,0);
}

void CircularViewRenderArea::paintEvent( QPaintEvent *e )
{
	if(needResizeView) {
		this->view->addUpdateFlags(GSLV_UF_ViewResized);
		needResizeView = false;
	}
	GSequenceLineViewRenderArea::paintEvent(e);
}

void CircularViewRenderArea::resizeEvent( QResizeEvent *e )
{
	//needResizeView = true;
	view->addUpdateFlags(GSLV_UF_ViewResized);
	QWidget::resizeEvent(e);
}
}//namespace

