/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "KalignTask.h"
#include "KalignAdapter.h"
#include "KalignConstants.h"
#include "TaskLocalStorage.h"

extern "C" {
#include "kalign2/kalign2_context.h"
}

#include <core_api/AppContext.h>
#include <core_api/AppSettings.h>
#include <core_api/AppResources.h>
#include <core_api/Log.h>
#include <core_api/StateLockableDataModel.h>
#include <core_api/DocumentModel.h>
#include <core_api/IOAdapter.h>
#include <core_api/Counter.h>
#include <gobjects/DNASequenceObject.h>
#include <util_tasks/LoadDocumentTask.h>

namespace GB2 {

static LogCategory log(ULOG_CAT_KALIGN);


void KalignTaskSettings::reset() {
	gapExtenstionPenalty = -1;
	gapOpenPenalty = -1;
	termGapPenalty = -1;
	secret = -1;
}

KalignTask::KalignTask(const MAlignment& ma, const KalignTaskSettings& _config) 
:Task(tr("KALIGN alignment"), TaskFlags_FOSCOE), config(_config), inputMA(ma)
{
	GCOUNTER( cvar, tvar, "KalignTask" );

	//int aliLen = ma.getLength();
	//int nSeq = ma.getNumSequences();

	ctx = new kalign_context();
	init_context(ctx, &stateInfo);
	if(config.gapOpenPenalty != -1) {
		ctx->gpo = config.gapOpenPenalty;
	}
	if(config.gapExtenstionPenalty != -1) {
		ctx->gpe = config.gapExtenstionPenalty;
	}
	if(config.termGapPenalty != -1) {
		ctx->tgpe = config.termGapPenalty;
	}
	if(config.secret != -1) {
		ctx->secret = config.secret;
	}
	//int memUseMB = int(quint64(200 * aliLen * nSeq) / (1024 * 1024)); //200x per char in alignment
	//TaskResourceUsage tru(RESOURCE_MEMORY, memUseMB);

	inputSubMA = inputMA;
	resultSubMA.alphabet = inputSubMA.alphabet;
	//if (config.alignRegion && config.regionToAlign.len != inputMA.getLength()) {
	//	assert(config.regionToAlign.len > 0);
	//	inputSubMA = inputMA.subAlignment(config.regionToAlign.startPos, config.regionToAlign.len);
	//}

	//ctx->tmp_uIds = new unsigned[inputSubMA.getNumSequences()];
	//for(unsigned i=0, n=inputSubMA.getNumSequences(); i<n; i++)
	//	ctx->input_uIds[i] = i;

	//if (config.nThreads == 1 || (config.op != KalignTaskOp_Align)) {
	tpm = Task::Progress_Manual;
	//} else {
	//	setUseDescriptionFromSubtask(true);
	//	parallelSubTask = new KalignParallelTask(inputSubMA, resultSubMA, config, ctx);
	//	addSubTask(parallelSubTask);
	//	tru.prepareStageLock = true;
	//}
	//taskResources.append(tru);
}

void KalignTask::run() {
	TaskLocalData::bindToKalignTLSContext(ctx);

	assert(!hasErrors());

	doAlign(); 

	if (!hasErrors() && !isCanceled()) {
		assert(resultMA.alphabet!=NULL);
		assert(resultMA.isNormalized());
	}
	TaskLocalData::detachKalignTLSContext();
}

void KalignTask::doAlign() {
	assert(resultSubMA.isEmpty());
	KalignAdapter::align(inputSubMA, resultSubMA, stateInfo);
	if (hasErrors()) {
		return;
	}
	resultMA = resultSubMA;
}

Task::ReportResult KalignTask::report() {
	delete ctx;
	ctx = NULL;
	return ReportResult_Finished;
}

//////////////////////////////////////////////////////////////////////////
// KalignGObjectTask

KalignGObjectTask::KalignGObjectTask(MAlignmentObject* _obj, const KalignTaskSettings& _config) 
: Task("", TaskFlags_NR_FOSCOE), obj(_obj), lock(NULL), kalignTask(NULL), config(_config)
{
	QString aliName = obj->getDocument()->getName();
	QString tn;
	tn = tr("KALIGN align '%1'").arg(aliName);
	setTaskName(tn);
	setUseDescriptionFromSubtask(true);
	setVerboseLogMode(true);
}

KalignGObjectTask::~KalignGObjectTask() {
	assert(lock == NULL);
}

void KalignGObjectTask::prepare() {
	if (obj.isNull()) {
		stateInfo.setError(tr("object_removed"));
		return;
	}
	if (obj->isStateLocked()) {
		stateInfo.setError(tr("object_is_state_locked"));
		return;
	}

	lock = new StateLock("kalign_lock");
	obj->lockState(lock);
	kalignTask = new KalignTask(obj->getMAlignment(), config);

	addSubTask(kalignTask);
}

Task::ReportResult KalignGObjectTask::report() {
	if (lock!=NULL) {
		obj->unlockState(lock);
		delete lock;
		lock = NULL;
	}
	propagateSubtaskError();
	if (hasErrors() || isCanceled()) {
		return ReportResult_Finished;
	}
	assert(!obj.isNull());
	if (obj->isStateLocked()) {
		stateInfo.setError(tr("object_is_state_locked"));
		return ReportResult_Finished;
	}
	assert(kalignTask->inputMA.getNumSequences() == kalignTask->resultMA.getNumSequences());
	obj->setMAlignment(kalignTask->resultMA);    

	return ReportResult_Finished;
}

} //namespace
